/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @author Stan Tomov
       @author Hartwig Anzt

       @generated from zcg_mic.cpp normal z -> c, Sun Jul 12 17:46:28 2015
*/

#include "magmasparse.h"
#include "magma.h"

#include <mkl.h>
#include <assert.h>

#define RTOLERANCE     slamch( "E" )
#define ATOLERANCE     slamch( "E" )

/**
    Purpose
    -------

    Solves a system of linear equations
       A * X = B
    where A is a complex Hermitian N-by-N positive definite matrix A.
    This is a GPU implementation of the Conjugate Gradient method.

    Arguments
    ---------

    @param[in]
    n           Problem size

    @param[in]
    daval       magmaFloatComplex_const_ptr
                Array of values for the input matrix A in CSR format

    @param[in]
    darow       magmaIndex_const_ptr
                Array of pointers to the rows of the input matrix A in CSR format

    @param[in]
    dacol       magmaIndex_const_ptr
                Array of column indices of the input matrix A in CSR format

    @param[in]
    db          magmaFloatComplex_const_ptr
                RHS b

    @param[in,out]
    dx          magmaFloatComplex_ptr
                solution approximation

    @param[in,out]
    dsolver_par magma_c_solver_par*
                solver parameters

    @param[in,out]                                                                               
    dwork       magmaFloatComplex_ptr
                work space of size 3 n

    @ingroup magmasparse_chesv
    ********************************************************************/

extern "C" magma_int_t
magma_ccg_mic(
    magma_int_t n, magmaFloatComplex_const_ptr daval, magmaIndex_const_ptr darow,
    magmaIndex_const_ptr dacol, magmaFloatComplex_const_ptr db, magmaFloatComplex_ptr dx,
    magma_c_solver_par *dsolver_par, magmaFloatComplex_ptr dwork)
{
    // prepare solver feedback
    dsolver_par->solver = Magma_CG;
    dsolver_par->numiter = 0;
    dsolver_par->info = MAGMA_SUCCESS; 

    magma_int_t dofs = n;

    // Prepare workspace vectors
    magmaFloatComplex_ptr r = dwork, d = dwork + dofs, z = dwork + 2*dofs;
    
    // solver variables
    magmaFloatComplex alpha, beta;
    float nom, nom0, r0, betanom, betanomsq, den, res;

    // solver setup
    nom = 0.;
    #pragma omp parallel for schedule(static,16) reduction(+:nom)
    for(int j=0; j<dofs; j++){
        dx[j] = MAGMA_C_ZERO;                                         //    x = 0
        r[j] = d[j] = db[j];                                          //    r = d = b
        nom += MAGMA_C_REAL(MAGMA_C_MUL(r[j], MAGMA_C_CNJG(r[j])));   //  nom = r' r
    }
    nom0 = sqrt(nom);                                                 // nom0 = sqrt(r' r)

    mkl_ccsrgemv("No Transpose", &dofs, (MKL_Complex8 *)daval,       //    z = A r
                 (MKL_INT *)darow, (MKL_INT *)dacol, 
                 (MKL_Complex8 *)r, (MKL_Complex8 *)z);

    den = 0.0;
    #pragma omp parallel for schedule(static,16) reduction(+:den)     
    for(int j=0; j<dofs; j++)
        den += MAGMA_C_REAL(MAGMA_C_MUL(z[j], MAGMA_C_CNJG(r[j])));   // den = r' z

    dsolver_par->init_res = nom0;
    if ( (r0 = nom * dsolver_par->epsilon) < ATOLERANCE ) {
        r0 = ATOLERANCE;
    }
    if ( nom < r0 ) {
        goto endpragma;
    }

    // check positive definite
    if (den <= 0.0) {
        printf("Operator A is not postive definite. (Ar,r) = %f\n", den);
        dsolver_par->info = MAGMA_NONSPD;
        goto endpragma;
    }

    //Chronometry
    real_Double_t tempo1, tempo2;
    tempo1 = dsecnd();
    if ( dsolver_par->verbose > 0 ) {
        dsolver_par->res_vec[0] = (real_Double_t)nom0;
        dsolver_par->timing[0] = 0.0;
    }
    
    // start iteration
    for( dsolver_par->numiter= 1; dsolver_par->numiter<dsolver_par->maxiter; 
                                                    dsolver_par->numiter++ ) {
        alpha = MAGMA_C_MAKE(nom/den, 0.);

        betanom = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:betanom)
        for(int j=0;j<dofs; j++){
            dx[j] += alpha*d[j];                                     // x = x + alpha d
            r[j]  -= alpha*z[j];                                     // r = r - alpha z
            betanom+=MAGMA_C_REAL(MAGMA_C_MUL(r[j], 
                                              MAGMA_C_CNJG(r[j])));  // betanom = r' * r
        }
        res = sqrt(betanom);

        if ( dsolver_par->verbose > 0 ) {
            tempo2 = dsecnd();
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) res;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) tempo2-tempo1;
            }
        }

        // fprintf(stderr,"Iteration %3d, scaled residual %e\n", dsolver_par->numiter, res/nom0);
        if (  res/nom0  < dsolver_par->epsilon ) {
            break;
        }

        beta = MAGMA_C_MAKE(betanom/nom, 0.);                        // beta = betanoms/nom
        #pragma omp parallel for schedule(static,16)
        for(int j=0; j<dofs; j++)                                    // d = r + beta * d
            d[j] = r[j] + beta * d[j];

        mkl_ccsrgemv("No Transpose", &dofs, (MKL_Complex8 *)daval,  // z = A d
                     (MKL_INT *)darow, (MKL_INT *)dacol, 
                     (MKL_Complex8 *)d, (MKL_Complex8 *)z);

        den = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:den)
        for(int j=0;j<dofs; j++)                                     // den = d' z
            den += MAGMA_C_REAL(MAGMA_C_MUL(z[j],
                                            MAGMA_C_CNJG(d[j])));

        nom = betanom;
    }
 
    tempo2 = dsecnd();

    dsolver_par->runtime = (real_Double_t) tempo2-tempo1;
    float residual = 0.;
    mkl_ccsrgemv("No Transpose", &dofs, (MKL_Complex8 *)daval,     // r = A x
                 (MKL_INT *)darow, (MKL_INT *)dacol,
                 (MKL_Complex8 *)dx, (MKL_Complex8 *)r);
    
    #pragma omp parallel for schedule(static,16) reduction(+:residual)
    for(int j=0; j<dofs; j++){
        r[j] = db[j] - r[j];
        residual+=MAGMA_C_REAL(MAGMA_C_MUL(r[j],                     // residual = r' r
                                           MAGMA_C_CNJG(r[j])));
    }
    dsolver_par->iter_res = res;
    dsolver_par->final_res = sqrt(residual);

    if ( dsolver_par->numiter < dsolver_par->maxiter) {
        dsolver_par->info = MAGMA_SUCCESS;
    } else if ( dsolver_par->init_res > dsolver_par->final_res ) {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_SLOW_CONVERGENCE;
    }
    else {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                    = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_DIVERGENCE;
    }

    endpragma: ;
     
    return MAGMA_SUCCESS;
}   /* magma_ccg_mic */
