/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @author Stan Tomov
       @author Hartwig Anzt 

       @generated from zbicgstab_mic.cpp normal z -> s, Sun Jul 12 17:46:28 2015
*/

#include "magmasparse.h"
#include "magma.h"

#include <mkl.h>
#include <assert.h>

#define RTOLERANCE     slamch( "E" )
#define ATOLERANCE     slamch( "E" )

/**
    Purpose
    -------

    Solves a system of linear equations
       A * X = B
    where A is a real symmetric N-by-N positive definite matrix A.
    This is a GPU implementation of the Biconjugate Gradient Stabelized method.

    Arguments
    ---------

    @param[in]
    n           Problem size

    @param[in]
    daval       magmaFloat_const_ptr
                Array of values for the input matrix A in CSR format

    @param[in]
    darow       magmaIndex_const_ptr
                Array of pointers to the rows of the input matrix A in CSR format

    dacol       magmaIndex_const_ptr
                Array of column indices of the input matrix A in CSR format

    @param[in]
    db          magmaFloat_const_ptr
                RHS b

    @param[in,out]
    dx          magmaFloat_ptr
                solution approximation

    @param[in,out]
    dsolver_par magma_s_solver_par*
                solver parameters

    @param[in,out]
    dwork       magmaFloat_ptr
                work space of size 6 n

    @ingroup magmasparse_sgesv
    ********************************************************************/

extern "C" magma_int_t
magma_sbicgstab_mic(
    magma_int_t n, magmaFloat_const_ptr daval, magmaIndex_const_ptr darow,
    magmaIndex_const_ptr dacol, magmaFloat_const_ptr db, magmaFloat_ptr dx,
    magma_s_solver_par *dsolver_par, magmaFloat_ptr dwork)
{
    // prepare solver feedback
    dsolver_par->solver = Magma_BICGSTAB;
    dsolver_par->numiter = 0;
    dsolver_par->info = MAGMA_SUCCESS;

    // some useful variables
    float c_zero = MAGMA_S_ZERO, c_one = MAGMA_S_ONE, 
                                            c_mone = MAGMA_S_NEG_ONE;
    
    magma_int_t dofs = n;

    // Prepare workspace vectors
    magmaFloat_ptr r = dwork, rr = dwork + n, p = dwork + 2*n;
    magmaFloat_ptr v = dwork + 3*n, s = dwork + 4*n, t = dwork + 5*n;
    
    // solver variables
    float alpha, beta, omega, rho_old, rho_new, val;
    float nom, betanom, nom0, r0, den, res, v1, v2;

    // solver setup
    nom = 0.;
    #pragma omp parallel for schedule(static,16) reduction(+:nom)
    for(int j=0; j<dofs; j++){
        dx[j] = MAGMA_S_ZERO;                                         //    x = 0
        r[j] = rr[j] = db[j];                                         //    r = rr = b
        nom += MAGMA_S_REAL(MAGMA_S_MUL(r[j], MAGMA_S_CNJG(r[j])));   //  nom = r' r
    }
    nom0 = betanom = sqrt(nom); 

    rho_old = omega = alpha = MAGMA_S_MAKE( 1.0, 0. );
    dsolver_par->init_res = nom0;

    mkl_scsrgemv("No Transpose", &dofs, (float *)daval,       //    v = A r
                 (MKL_INT *)darow, (MKL_INT *)dacol,
                 (float *)r, (float *)v);

    den = 0.;
    #pragma omp parallel for schedule(static,16) reduction(+:den)
    for(int j=0; j<dofs; j++)
        den += MAGMA_S_REAL(MAGMA_S_MUL(r[j], MAGMA_S_CNJG(v[j])));   // den = v' r 

    if ( (r0 = nom * dsolver_par->epsilon) < ATOLERANCE ) {
        r0 = ATOLERANCE;
    }
    if ( nom < r0 ) {
        return MAGMA_SUCCESS;
    }
    // check positive definite  
    if (den <= 0.0) {
        printf("Operator A is not postive definite. (Ar,r) = %f\n", den);
        dsolver_par->info = MAGMA_NONSPD;
        return MAGMA_NONSPD;
    }

    //Chronometry
    real_Double_t tempo1, tempo2;
    tempo1 = dsecnd();
    if ( dsolver_par->verbose > 0 ) {
        dsolver_par->res_vec[0] = nom0;
        dsolver_par->timing[0] = 0.0;
    }

    // start iteration
    for( dsolver_par->numiter= 1; dsolver_par->numiter<dsolver_par->maxiter; 
                                                    dsolver_par->numiter++ ) {

        v1 = v2 = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:v1,v2)
        for(int j=0; j<dofs; j++){
            val = MAGMA_S_MUL(MAGMA_S_CNJG(rr[j]), r[j]);     //  rho = rr' r                
            v1 += MAGMA_S_REAL(val);
            v2 += MAGMA_S_IMAG(val);
        }
        rho_new = MAGMA_S_MAKE(v1, v2);
        beta = MAGMA_S_MUL( MAGMA_S_DIV(rho_new, rho_old),
                            MAGMA_S_DIV(alpha  ,   omega));   // beta=rho/rho_old *alpha/omega

        val = MAGMA_S_MUL( beta, omega);
        #pragma omp parallel for schedule(static,16)          // p = beta*p - omega*beta*v + r
        for(int j=0; j<dofs; j++){
            p[j] = MAGMA_S_MUL(beta, p[j]) - MAGMA_S_MUL(val, v[j]) + r[j];
        }

        mkl_scsrgemv("No Transpose", &dofs,                   // v = Ap 
                     (float *)daval, (MKL_INT *)darow, (MKL_INT *)dacol,
                     (float *)p, (float *)v );

        v1 = v2 = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:v1,v2)
        for(int j=0; j<dofs; j++){
            val = MAGMA_S_MUL(MAGMA_S_CNJG(rr[j]), v[j]);     //  rho = rr' v
            v1 += MAGMA_S_REAL(val);
            v2 += MAGMA_S_IMAG(val);
        }
        alpha = MAGMA_S_DIV(rho_new, MAGMA_S_MAKE(v1, v2));

        #pragma omp parallel for schedule(static,16)
        for(int j=0; j<dofs; j++){                            // s = r - alpha*v
            s[j] = r[j] - MAGMA_S_MUL(alpha, v[j]);
        }
        
        mkl_scsrgemv("No Transpose", &dofs,                   // t = A s
                     (float *)daval, (MKL_INT *)darow, (MKL_INT *)dacol,
                     (float *)s, (float *)t );

        v1 = v2 = res = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:v1,v2,res)
        for(int j=0; j<dofs; j++){
            val = MAGMA_S_MUL(MAGMA_S_CNJG(t[j]), s[j]);      // val = t' s
            v1 += MAGMA_S_REAL(val);
            v2 += MAGMA_S_IMAG(val);
            res += ( MAGMA_S_REAL(t[j]) * MAGMA_S_REAL(t[j]) +
                     MAGMA_S_IMAG(t[j]) * MAGMA_S_IMAG(t[j])   );
        }
        omega = MAGMA_S_MAKE(v1, v2) / res;                  // omega = t' s / <t,t>
        
        nom = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:nom)
        for(int j=0; j<dofs; j++){
            dx[j] += ( MAGMA_S_MUL(alpha, p[j]) +            // x=x+alpha*p + omega*s
                       MAGMA_S_MUL(omega, s[j])  );
            r[ j]  = s[j] - MAGMA_S_MUL(omega, t[j]);        // r = s - omega * t
            nom += ( MAGMA_S_REAL(r[j]) * MAGMA_S_REAL(r[j]) +
                     MAGMA_S_IMAG(r[j]) * MAGMA_S_IMAG(r[j])   );
        }
        res = betanom = sqrt(nom);
        rho_old = rho_new;                                   // rho_old=rho
        if ( dsolver_par->verbose > 0 ) {
            tempo2 = dsecnd();
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) res;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }

        // fprintf(stderr,"Iteration %3d, scaled residual %e\n", dsolver_par->numiter, res/nom0);
        if ( res/nom0  < dsolver_par->epsilon ) {
            break;
        }
    }

    tempo2 = dsecnd();
    dsolver_par->runtime = (real_Double_t) tempo2-tempo1;
    float residual = 0.;
    mkl_scsrgemv("No Transpose", &dofs, (float *)daval,     // r = A x
                 (MKL_INT *)darow, (MKL_INT *)dacol,
                 (float *)dx, (float *)r);

    #pragma omp parallel for schedule(static,16) reduction(+:residual)
    for(int j=0; j<dofs; j++){
        r[j] = db[j] - r[j];
        residual += ( MAGMA_S_REAL(r[j])*MAGMA_S_REAL(r[j]) +       // residual = r' r
                      MAGMA_S_IMAG(r[j])*MAGMA_S_IMAG(r[j])  );
    }
    dsolver_par->iter_res = betanom;
    dsolver_par->final_res = sqrt(residual);

    if ( dsolver_par->numiter < dsolver_par->maxiter) {
        dsolver_par->info = MAGMA_SUCCESS;
    } else if ( dsolver_par->init_res > dsolver_par->final_res ) {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_SLOW_CONVERGENCE;
    }
    else {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_DIVERGENCE;
    }

    return MAGMA_SUCCESS;
}   /* magma_sbicgstab */


