/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @author Stan Tomov
       @author Hartwig Anzt

       @precisions normal z -> s d c
*/

#include "magmasparse.h"
#include "magma.h"

#include <mkl.h>
#include <assert.h>

#define PRECISION_z

#define  q(i,j)    (q + (i)*dofs + (j))
#define  H(i,j)  H[(i)   + (j)*(1+ldh)]
#define HH(i,j) HH[(i)   + (j)*ldh]

#define min(a, b) ((a) < (b) ? (a) : (b))

#define RTOLERANCE     dlamch( "E" )
#define ATOLERANCE     dlamch( "E" )

/**
    Purpose
    -------

    Solves a system of linear equations
       A * X = B
    where A is a complex sparse matrix stored in the GPU memory.
    X and B are complex vectors stored on the GPU memory. 
    This is a GPU implementation of the GMRES method.

    Arguments
    ---------

    @param[in]
    A           magma_z_sparse_matrix
                descriptor for matrix A

    @param[in]
    b           magma_z_vector
                RHS b vector

    @param[in,out]
    x           magma_z_vector*
                solution approximation

    @param[in,out]
    solver_par  magma_z_solver_par*
                solver parameters

    @param[in,out]
    dwork       magmaDoubleComplex_ptr
                work space of size 
                (dsolver_par->restart + 3) n + (dsolver_par->restart + 1) (5+2*dsolver_par->restart)

    @ingroup magmasparse_zgesv
    ********************************************************************/

extern "C" magma_int_t
magma_zgmres_mic(
    magma_int_t n, magmaDoubleComplex_const_ptr daval, magmaIndex_const_ptr darow,
    magmaIndex_const_ptr dacol, magmaDoubleComplex_const_ptr db, magmaDoubleComplex_ptr dx,
    magma_z_solver_par *dsolver_par, magmaDoubleComplex_ptr dwork)                 
{
    magma_int_t stat = 0;    
    magma_int_t stat_cpu = 0, stat_dev = 0;

    // prepare solver feedback
    dsolver_par->solver = Magma_GMRES;
    dsolver_par->numiter = 0;
    dsolver_par->info = MAGMA_SUCCESS;

    // local variables
    magmaDoubleComplex c_zero = MAGMA_Z_ZERO, c_one = MAGMA_Z_ONE, c_mone = MAGMA_Z_NEG_ONE;
    int ione = 1;
    magma_int_t dofs = n;
    magma_int_t i, j, k, m = 0;
    magma_int_t restart = min( dofs-1, dsolver_par->restart );
    magma_int_t ldh = restart+1;
    double nom, rNorm, RNorm, nom0, betanom, r0 = 0.;

    // Prepare workspace vectors
    magmaDoubleComplex *r, *q, *q_t, *H, *HH, *y, *h1, v;
    r  = dwork;
    q  = r + dofs; 
    H  = q + dofs *(ldh+1);
    y  = H  + (ldh+1)*ldh;
    HH = y  + ldh;
    h1 = HH + ldh*ldh;

    nom = 0.;
    #pragma omp parallel for schedule(static,16) reduction(+:nom)
    for(int j=0; j<dofs; j++){
        dx[j] = MAGMA_Z_ZERO;                                     //    x = 0
        r[ j] = db[j];                                            //    r = b
        nom+=MAGMA_Z_REAL(MAGMA_Z_MUL(r[j], MAGMA_Z_CNJG(r[j]))); //  nom = r' r
    }
    nom0 = betanom = sqrt(nom);                                   // nom0 = ||r||

    dsolver_par->init_res = nom0;

    H(1,0) = MAGMA_Z_MAKE( nom0, 0. );
    if ( (r0 = nom0 * dsolver_par->epsilon ) < ATOLERANCE ){ 
        r0 = dsolver_par->epsilon;
    }

    if ( nom < r0 ) {
        return MAGMA_SUCCESS;
    }

    //Chronometry
    real_Double_t tempo1, tempo2;
    tempo1 = dsecnd();
    if ( dsolver_par->verbose > 0 ) {
        dsolver_par->res_vec[0] = nom0;
        dsolver_par->timing[0] = 0.0;
    }

    // start iteration
    for( dsolver_par->numiter= 1; dsolver_par->numiter<dsolver_par->maxiter; 
         dsolver_par->numiter++ ) {

        for(k=1; k<=restart; k++) {
            v = MAGMA_Z_DIV(MAGMA_Z_ONE, H(k,k-1));

            #pragma omp parallel for schedule(static,16)
            for(int j=0; j<dofs; j++){                                     //  q(k-1) = r / H(k,k-1)
                *q(k-1,j) = MAGMA_Z_MUL(v, r[j]);
            } 
            
            mkl_zcsrgemv("No Transpose", &dofs, (MKL_Complex16 *)daval,    //       r = A q(k-1)
                         (MKL_INT *)darow, (MKL_INT *)dacol,
                         (MKL_Complex16 *)q(k-1,0), (MKL_Complex16 *)r);            

            // modified Gram-Schmidt
            for (i=1; i<=k; i++) {
                double v1 = 0., v2 = 0.;
                #pragma omp parallel for schedule(static,16) reduction(+:v1,v2)
                for(int j=0; j<dofs; j++){
                    v = MAGMA_Z_MUL(MAGMA_Z_CNJG(*q(i-1,j)), r[j]);       //  H(i,k) = q[i]' r 
                    v1 += MAGMA_Z_REAL(v);
                    v2 += MAGMA_Z_IMAG(v);
                }
                H(i, k) = MAGMA_Z_MAKE(v1, v2);
                
                #pragma omp parallel for schedule(static,16)
                for(int j=0; j<dofs; j++){
                    r[j] -= MAGMA_Z_MUL(H(i,k), *q(i-1,j));               //  r = r - H(i,k) q[i]
                }
            }

            nom = 0.;
            #pragma omp parallel for schedule(static,16) reduction(+:nom)
            for(int j=0; j<dofs; j++) {
                nom += MAGMA_Z_REAL(MAGMA_Z_MUL(r[j], MAGMA_Z_CNJG(r[j])));
            }
            H(k+1,k) = MAGMA_Z_MAKE( sqrt(nom), 0.);

            // Minimization of  || b-Ax ||  in H_k             
            for (i=1; i<=k; i++) {
                int ip = i+1;
                
                #if defined(PRECISION_z) || defined(PRECISION_c)
                zdotc( (MKL_Complex16 *)&HH(k,i), &ip, (MKL_Complex16 *)&H(1,k), &ione, 
                       (MKL_Complex16 *)&H(1,i), &ione );
                #else
                   HH(k,i) = zdotc( &ip, &H(1,k), &ione, &H(1,i), &ione );
                #endif
            }
            h1[k] = H(1,k)*H(1,0); 
            if (k != 1) {
                for (i=1; i<k; i++) {
                    HH(k,i) = HH(k,i)/HH(i,i);
                    for (m=i+1; m<=k; m++) {
                        HH(k,m) -= HH(k,i) * HH(m,i) * HH(i,i);
                    }
                    h1[k] -= h1[i] * HH(k,i);   
                }    
            }
        
        y[k] = h1[k]/HH(k,k); 
        if (k != 1)  
            for (i=k-1; i>=1; i--) {
                y[i] = h1[i]/HH(i,i);
                for (int j=i+1; j<=k; j++)
                    y[i] -= y[j] * HH(j,i);
            }                    
        m = k;
        rNorm = fabs(MAGMA_Z_REAL(H(k+1,k)));
        }
        /*     Minimization done       */ 

        // compute solution approximation
        #pragma omp parallel for schedule(static,16) 
        for(int j=0; j<dofs; j++) {
            magmaDoubleComplex val = dx[j];
            for (i=0; i<m; i++) {
                val += MAGMA_Z_MUL(y[i+1], *q(i,j)); 
            }  
            dx[j] = val;
        }

        // compute residual
        mkl_zcsrgemv("No Transpose", &dofs, (MKL_Complex16 *)daval,     // r = A x
                     (MKL_INT *)darow, (MKL_INT *)dacol,
                     (MKL_Complex16 *)dx, (MKL_Complex16 *)r);

        double residual = 0.;
        #pragma omp parallel for schedule(static,16) reduction(+:residual)
        for(int j=0; j<dofs; j++){
            r[j] = db[j] - r[j];
            residual+=MAGMA_Z_REAL(MAGMA_Z_MUL(r[j],                    // residual = r' r
                                               MAGMA_Z_CNJG(r[j])));
        }

        H(1,0) = MAGMA_Z_MAKE( sqrt(residual), 0. );                    //  RNorm = H[1][0] = || r ||
        RNorm = MAGMA_Z_REAL( H(1,0) );
        betanom = fabs(RNorm);  

        if ( dsolver_par->verbose > 0 ) {
            tempo2 = dsecnd();
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }

        // fprintf(stderr,"Iteration %3d, scaled residual %e\n", dsolver_par->numiter, betanom);
        if (  betanom  < r0 ) {
            break;
        } 
    }

    tempo2 = dsecnd();
    dsolver_par->runtime   = (real_Double_t) tempo2-tempo1;
    dsolver_par->iter_res  = betanom;
    dsolver_par->final_res = betanom;

    if ( dsolver_par->numiter < dsolver_par->maxiter) {
        dsolver_par->info = MAGMA_SUCCESS;
    } else if ( dsolver_par->init_res > dsolver_par->final_res ) {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_SLOW_CONVERGENCE;
    }
    else {
        if ( dsolver_par->verbose > 0 ) {
            if ( (dsolver_par->numiter)%dsolver_par->verbose==0 ) {
                dsolver_par->res_vec[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) betanom;
                dsolver_par->timing[(dsolver_par->numiter)/dsolver_par->verbose] 
                        = (real_Double_t) tempo2-tempo1;
            }
        }
        dsolver_par->info = MAGMA_DIVERGENCE;
    }

    return MAGMA_SUCCESS;
}   /* magma_zgmres */

