/*
    -- micMAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @author Ichitaro Yamazaki
       @author Stan Tomov      

       @precisions normal z -> s d c
*/

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

#include <immintrin.h>

#include "magmasparse.h"
#include "magma.h"
#include "magma_z_server.h"

#define min(a, b) ((a) < (b) ? (a) : (b))

#define PRECISION_z
#define COMPLEX


#define PRECISION_z

//==================================================================
#define max(a, b) ((a) > (b) ? (a) : (b))

void 
magma_zcopy_mic(
    magma_int_t n,
    magmaDoubleComplex_const_ptr dx_src, magma_int_t incx,
    magmaDoubleComplex_ptr       dy_dst, magma_int_t incy,
    magma_queue_t queue )
{
  if ( n <= 0 )
    return ;

  magma_int_t one =1;
  if (incx == 1 && incy == 1)
    zlacpy("All", &n, &one,
           (MKL_Complex16 *)dx_src, &n,
           (MKL_Complex16 *)dy_dst, &n);
  else
    zlacpy("All", &one, &n,
           (MKL_Complex16 *)dx_src, &incx,
           (MKL_Complex16 *)dy_dst, &incy );
}

extern "C" void
magma_zdscal_mic( 
    magma_int_t k, 
    double R1,
    magmaDoubleComplex_ptr dx, 
    magma_int_t incx, magma_queue_t queue)
{
    for(magma_int_t i =0; i<k; i++)
      dx[i*incx] *= R1;
}

//==================================================================
#define NB 64
#define A(i,j) (A[(i) + (j)*lda])
#define W(i,j) (W[(i) + (j)*ldw])

// each thread block does one NB x n block row of A.
// each thread does one row, starting from left edge and moving right to diagonal.
void
zlascl_2x2_lower(magma_int_t grid, magma_int_t threads, 
                 magma_int_t m, const magmaDoubleComplex* W, magma_int_t ldw, magmaDoubleComplex* A, magma_int_t lda)
{
    //#pragma omp parallel for
   for(magma_int_t blockIdx=0; blockIdx < grid; blockIdx++)
     {
       for(magma_int_t threadIdx=0; threadIdx < threads; threadIdx++)
	 { 
	   magma_int_t ind = blockIdx * 64 + threadIdx;

	   magmaDoubleComplex D21 = W( 1, 0 );
	   magmaDoubleComplex D11 = MAGMA_Z_DIV( W( 1, 1 ), D21 );
	   magmaDoubleComplex D22 = MAGMA_Z_DIV( W( 0, 0 ), MAGMA_Z_CNJG( D21 ) );
	   double T = 1.0 / ( MAGMA_Z_REAL( D11*D22 ) - 1.0 );
	   D21 = MAGMA_Z_DIV( MAGMA_Z_MAKE(T,0.0), D21 );
	 
	   if (ind < m) {
	     A( ind, 0 ) = MAGMA_Z_CNJG( D21 )*( D11*W( 2+ind, 0 )-W( 2+ind, 1 ) );
	     A( ind, 1 ) = D21*( D22*W( 2+ind, 1 )-W( 2+ind, 0 ) );
	   }
	 }
     }
}

// each thread block does one NB x n block row of A.
// each thread does one row, starting from right edge and moving left to diagonal.
void
zlascl_2x2_upper(magma_int_t grid, magma_int_t threads,
                 magma_int_t m, const magmaDoubleComplex *W, magma_int_t ldw, magmaDoubleComplex* A, magma_int_t lda)
{

    //#pragma omp parallel for
  for(magma_int_t blockIdx=0; blockIdx < grid; blockIdx++) 
    {
      for(magma_int_t threadIdx=0; threadIdx < threads; threadIdx++)
	{ 

	  magma_int_t ind = blockIdx * 64 + threadIdx;
	  
	  magmaDoubleComplex D21 = W( m, 1 );
	  magmaDoubleComplex D11 = MAGMA_Z_DIV( W( m+1, 1 ), MAGMA_Z_CNJG( D21 ) );
	  magmaDoubleComplex D22 = MAGMA_Z_DIV( W( m, 0 ), D21 );
	  double T = 1.0 / ( MAGMA_Z_REAL( D11*D22 ) - 1.0 );
	  D21 = MAGMA_Z_DIV( MAGMA_Z_MAKE(T,0.0), D21 );
	
	  if (ind < m) {
	    A( ind, 0 ) = D21*( D11*W( ind, 0 )-W( ind, 1 ) );
	    A( ind, 1 ) = MAGMA_Z_CNJG( D21 )*( D22*W( ind, 1 )-W( ind, 0 ) );
	  }
	  
	}
    }
}

#undef A
#undef W

extern "C" void
magmablas_zlascl_2x2( 
    magma_type_t type, magma_int_t m,
    const magmaDoubleComplex *dW, magma_int_t lddw,
    magmaDoubleComplex *dA, magma_int_t ldda,
    magma_int_t *info, magma_queue_t queue)
{
    *info = 0;
    if ( type != MagmaLower && type != MagmaUpper && type != MagmaFull ) {
       *info = -1;
    } else if ( m < 0 ) {
       *info = -2;
    } else if ( ldda < max(1,m) ) {
       *info = -4;
    }

    if (*info != 0) {
       magma_xerbla( __func__, -(*info) );
       return;  //info;
    }

    magma_int_t grid    = (m + NB - 1)/NB;
    magma_int_t threads = NB; 

    if (type == MagmaLower) {
      zlascl_2x2_lower( grid, threads, m, dW, lddw, dA, ldda);
    }
    else {
      zlascl_2x2_upper( grid, threads, m, dW, lddw, dA, ldda);
    }
}

//==================================================================

/*
 *  First version: line per line
 */
typedef struct {
  magmaDoubleComplex *A1;
  magmaDoubleComplex *A2;
  magma_int_t n, lda1, lda2;
} magmagpu_zlacpy_cnjg_params_t;

void magmagpu_zlacpy_cnjg(magma_int_t grid, magma_int_t threads, magmagpu_zlacpy_cnjg_params_t params )
{
   //#pragma omp parallel for
  for(magma_int_t blockIdx=0; blockIdx < grid; blockIdx++)
    {
      for(magma_int_t threadIdx=0; threadIdx < threads; threadIdx++)
	{

	  magma_int_t x = threadIdx + 64 * blockIdx;
	  magma_int_t offset1 = x*params.lda1;
	  magma_int_t offset2 = x*params.lda2;
	  if( x < params.n )
	    {
	      magmaDoubleComplex *A1  = params.A1 + offset1;
	      magmaDoubleComplex *A2  = params.A2 + offset2;
	      *A2 = MAGMA_Z_CNJG(*A1);
	    }
	  
	}
    }
}


void
magmablas_zlacpy_cnjg(
    magma_int_t n, magmaDoubleComplex *dA1, magma_int_t lda1,
    magmaDoubleComplex *dA2, magma_int_t lda2, magma_queue_t queue)
{
    int threads = 64;
    int blocks  =  (n+threads-1) / threads;
    magmagpu_zlacpy_cnjg_params_t params = { dA1, dA2, n, lda1, lda2 };
    magmagpu_zlacpy_cnjg(blocks, threads, params );
}

int izamax_2(magma_int_t *n, MKL_Complex16 *a, magma_int_t *inc){
    int res = izamax( n, a, inc );
    fprintf(stderr,"izamax = %d\n", res); 
    return res;
}

//==================================================================

/**
    Purpose
    =======

    ZLAHEF computes a partial factorization of a complex Hermitian
    matrix A using the Bunch-Kaufman diagonal pivoting method. The
    partial factorization has the form:

    A  =  ( I  U12 ) ( A11  0  ) (  I    0   )  if UPLO = 'U', or:
          ( 0  U22 ) (  0   D  ) ( U12' U22' )

    A  =  ( L11  0 ) (  D   0  ) ( L11' L21' )  if UPLO = 'L'
          ( L21  I ) (  0  A22 ) (  0    I   )

    where the order of D is at most NB. The actual order is returned in
    the argument KB, and is either NB or NB-1, or N if N <= NB.
    Note that U' denotes the conjugate transpose of U.

    ZLAHEF is an auxiliary routine called by ZHETRF. It uses blocked code
    (calling Level 3 BLAS) to update the submatrix A11 (if UPLO = 'U') or
    A22 (if UPLO = 'L').

    Arguments
    ---------
    @param[in]
    UPLO    CHARACTER
            Specifies whether the upper or lower triangular part of the
            Hermitian matrix A is stored:
      -     = 'U':  Upper triangular
      -     = 'L':  Lower triangular

    @param[in]
    N       INTEGER
            The order of the matrix A.  N >= 0.

    @param[in]
    NB      INTEGER
            The maximum number of columns of the matrix A that should be
            factored.  NB should be at least 2 to allow for 2-by-2 pivot
            blocks.

    @param[out]
    KB      INTEGER
            The number of columns of A that were actually factored.
            KB is either NB-1 or NB, or N if N <= NB.

    @param[in,out]
    A       COMPLEX*16 array, dimension (LDA,N)
            On entry, the Hermitian matrix A.  If UPLO = 'U', the leading
            n-by-n upper triangular part of A contains the upper
            triangular part of the matrix A, and the strictly lower
            triangular part of A is not referenced.  If UPLO = 'L', the
            leading n-by-n lower triangular part of A contains the lower
            triangular part of the matrix A, and the strictly upper
            triangular part of A is not referenced.
            On exit, A contains details of the partial factorization.

    @param[in]
    LDA     INTEGER
            The leading dimension of the array A.  LDA >= max(1,N).

    @param[out]
    ipiv    INTEGER array, dimension (N)
            Details of the interchanges and the block structure of D.
            If UPLO = 'U', only the last KB elements of ipiv are set;
            if UPLO = 'L', only the first KB elements are set.
    \n
            If ipiv(k) > 0, then rows and columns k and ipiv(k) were
            interchanged and D(k,k) is a 1-by-1 diagonal block.
            If UPLO = 'U' and ipiv(k) = ipiv(k-1) < 0, then rows and
            columns k-1 and -ipiv(k) were interchanged and D(k-1:k,k-1:k)
            is a 2-by-2 diagonal block.  If UPLO = 'L' and ipiv(k) =
            ipiv(k+1) < 0, then rows and columns k+1 and -ipiv(k) were
            interchanged and D(k:k+1,k:k+1) is a 2-by-2 diagonal block.
 
    @param[out]
    W       (workspace) COMPLEX*16 array, dimension (LDW,NB)
 
    @param[in]
    LDW     INTEGER
            The leading dimension of the array W.  LDW >= max(1,N).

    @param[out]
    INFO    INTEGER
      -     = 0: successful exit
      -     > 0: if INFO = k, D(k,k) is exactly zero.  The factorization
                 has been completed, but the block diagonal matrix D is
                 exactly singular.
  
    @ingroup magma_zhetrf_comp
    ********************************************************************/

void call_zlahef(magma_mic_zlahef_param param)
{
    /*  Arguments passed from the CPU  */
    magma_uplo_t uplo         = param.uplo;
    magma_int_t  n            = param.n;
    magma_int_t nb            = param.nb;
    magma_int_t *kb           = param.kb;
    magmaDoubleComplex *hA    = param.hA;
    magma_int_t lda           = param.lda;
    magmaDoubleComplex_ptr dA = param.dA;
    magma_int_t ldda          = param.ldda;
    magma_int_t *ipiv         = param.ipiv;
    magmaDoubleComplex_ptr dW = param.dW;
    magma_int_t lddw          = param.lddw;
    magma_int_t info;
    magma_queue_t queue;

    /* .. Parameters .. */
    double d_one   = 1.0;
    double d_zero  = 0.0;
    double d_eight = 8.0;
    double d_seven = 7.0;
    #if defined(PRECISION_c)
    float  f_zero =  0.0;
    #endif
    magmaDoubleComplex c_one  =  MAGMA_Z_ONE;
    magmaDoubleComplex c_mone = -MAGMA_Z_ONE;
    magma_int_t upper = (uplo == MagmaUpper);
    magma_int_t ione = 1, lm, ln, lk;
  
    /* .. Local Scalars .. */
    magma_int_t imax = 0, jmax = 0, kk, kkW, kp, kstep, iinfo;
    double   abs_akk, alpha, colmax, R1, rowmax;
    magmaDoubleComplex Zimax, Z;

    #define dA(i, j)  (dA + (j)*ldda  + (i))
    #define dW(i, j)  (dW + (j)*lddw  + (i))
    #define  A(i, j)  (hA + (j)*lda   + (i))

    /* .. Executable Statements .. */
    info = 0;

    /* Initialize alpha for use in choosing pivot block size. */
    alpha = ( d_one+sqrt( d_seven ) ) / d_eight;

    magma_event_t event = NULL;
    if( upper ) {
       /* Factorize the trailing columns of A using the upper triangle
          of A and working backwards, and compute the matrix W = U12*D
          for use in updating A11 (note that conjg(W) is actually stored)
 
          K is the main loop index, decreasing from N in steps of 1 or 2
 
          KW is the column of W which corresponds to column K of A   */
       magma_int_t k, kw = 0;
       for (k = n-1; k+1 > max(n-nb+1, nb); k -= kstep) {
           kw = nb - (n-k);
           /* Copy column K of A to column KW of W and update it */

           magma_zcopy_mic( k+1, dA( 0, k ), 1, dW( 0, kw ), 1, queue );

           // set imaginary part of diagonal to be zero
           #if defined(PRECISION_z) || defined(PRECISION_c)
           MAGMA_Z_IMAG( *dW( k, kw ) ) = 0.;
           #endif

           if (k+1 < n) {
                lm = k+1;
                ln = n-(k+1); 
                zgemv( lapack_const(MagmaNoTrans), &lm, &ln, 
                       (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( 0, k+1 ), &ldda,
                       (MKL_Complex16 *)dW( k, kw+1 ), &lddw, (MKL_Complex16 *)&c_one, 
                       (MKL_Complex16 *)dW( 0, kw ), &ione);

                // set imaginary part of diagonal to be zero
                #if defined(PRECISION_z) || defined(PRECISION_c)
                MAGMA_Z_IMAG( *dW( k, kw ) ) = 0.;
                #endif
           }

           kstep = 1;

           /* Determine rows and columns to be interchanged and whether
              a 1-by-1 or 2-by-2 pivot block will be used */
           abs_akk = fabs( MAGMA_Z_REAL( *dW( k, kw ) ) );

           /* imax is the row-index of the largest off-diagonal element in
              column K, and colmax is its absolute value */
           if( k > 0 ) {
                // magma is one-base
                imax = izamax( &k, (MKL_Complex16 *)dW( 0, kw ), &ione ) - 1;
                colmax = MAGMA_Z_ABS1( *dW( imax, kw ) );
           } else {
                colmax = d_zero;
           }
           if( max( abs_akk, colmax ) == 0.0 ) {
            
                /* Column K is zero: set INFO and continue */
                if ( info == 0 ) info = k;

                kp = k;

                #if defined(PRECISION_z) || defined(PRECISION_c)
                MAGMA_Z_IMAG( *dA( k, k ) ) = 0.;
                #endif
           } else {
            if( abs_akk >= alpha*colmax ) {

              /* no interchange, use 1-by-1 pivot block */
              kp = k;
            } else {

              /* Copy column imax to column KW-1 of W and update it */
              magma_zcopy_mic( imax+1, dA( 0, imax ), 1, dW( 0, kw-1 ), 1, queue );
              #if defined(PRECISION_z) || defined(PRECISION_c)
              MAGMA_Z_IMAG( *dW( imax, kw-1 ) ) = 0.;
              #endif

              #if defined(PRECISION_z) || defined(PRECISION_c)
              magmablas_zlacpy_cnjg( k-imax, dA(imax,imax+1), ldda, dW(imax+1,kw-1), 1, queue );
              #else
              magma_zcopy_mic( k-imax, dA(imax,imax+1), ldda, dW(imax+1,kw-1), 1, queue );
              #endif
              if( k+1 < n ) {
                lm = k+1;
                ln = n-(k+1);
		zgemv( lapack_const(MagmaNoTrans), &lm, &ln, 
                       (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( 0, k+1 ), &ldda, 
                       (MKL_Complex16 *)dW( imax, kw+1 ), &lddw,
                       (MKL_Complex16 *)&c_one, (MKL_Complex16 *)dW( 0, kw-1 ), &ione );

                 #if defined(PRECISION_z) || defined(PRECISION_c)
                 MAGMA_Z_IMAG( *dW( imax, kw-1 ) ) = 0.;
                 #endif
              }
              Zimax = *dW( imax, kw-1 );

              /* jmax is the column-index of the largest off-diagonal
                element in row imax, and rowmax is its absolute value */
              lm = k-imax;
              jmax = imax + izamax( &lm, (MKL_Complex16 *)dW( imax+1, kw-1 ), &ione );
              rowmax = MAGMA_Z_ABS1( *dW( jmax, kw-1 ) );
              if ( imax > 0 ) {
               // magma is one-base
		jmax = izamax( &imax, (MKL_Complex16 *)dW( 0, kw-1 ), &ione ) - 1;
		rowmax = max( rowmax, MAGMA_Z_ABS1( *dW( jmax, kw-1 ) ) );
              }

              if( abs_akk >= alpha*colmax*( colmax / rowmax ) ) {
                     /* no interchange, use 1-by-1 pivot block */
                     kp = k;
              } else if ( fabs( MAGMA_Z_REAL( Zimax ) ) >= alpha*rowmax ) {
                     /* interchange rows and columns K and imax, use 1-by-1
                        pivot block */
                     kp = imax;

                     /* copy column KW-1 of W to column KW */
                     magma_zcopy_mic( k+1, dW( 0, kw-1 ), 1, dW( 0, kw ), 1, queue );
                } else {

                     /* interchange rows and columns K-1 and imax, use 2-by-2
                        pivot block */
                     kp = imax;
                     kstep = 2;
                }
           }
           kk = k - kstep + 1;
           kkW = nb - (n - kk);

           /* Updated column kp is already stored in column kkW of W */
           if( kp != kk ) {

             /* Interchange rows kk and kp in last kk columns of A and W */
             // note: row-swap A(:,kk)
             ln = n-kk;
             zswap( &ln, (MKL_Complex16 *)dA( kk, kk ), &ldda, (MKL_Complex16 *)dA( kp, kk ), &ldda );
             zswap( &ln, (MKL_Complex16 *)dW( kk, kkW), &lddw, (MKL_Complex16 *)dW( kp, kkW), &lddw );

             /* Copy non-updated column kk to column kp */
             #if defined(PRECISION_z) || defined(PRECISION_c)
             magmablas_zlacpy_cnjg( kk-kp-1, dA( kp+1, kk ), 1, dA( kp, kp+1 ), ldda, queue );
             #else
             magma_zcopy_mic( kk-kp-1, dA( kp+1, kk ), 1, dA( kp, kp+1 ), ldda, queue );
             #endif

             // now A(kp,kk) should be A(kk,kk), and copy to A(kp,kp)
             magma_zcopy_mic( kp+1, dA( 0, kk ), 1, dA( 0, kp ), 1, queue );
             #if defined(PRECISION_z) || defined(PRECISION_c)
             MAGMA_Z_IMAG( *dA( kp, kp ) ) = 0.;
             #endif
           }
           if( kstep == 1 ) {

                /* 1-by-1 pivot block D(k): column KW of W now holds
                      W(k) = U(k)*D(k)
                      where U(k) is the k-th column of U 
                      Store U(k) in column k of A */
                magma_zcopy_mic( k+1, dW( 0, kw ), 1, dA( 0, k ), 1, queue );
                if ( k > 0 ) {
                   R1 = d_one / MAGMA_Z_REAL( *dA( k, k ) );
                   magma_zdscal_mic( k, R1, dA( 0, k ), 1, queue );

                   /* Conjugate W(k) */
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   magmablas_zlacpy_cnjg( k, dW( 0, kw ),1, dW( 0, kw ),1, queue );
                   #endif
                }
           } else {

            /* 2-by-2 pivot block D(k): columns KW and KW-1 of W now hold
              ( W(k-1) W(k) ) = ( U(k-1) U(k) )*D(k)
              where U(k) and U(k-1) are the k-th and (k-1)-th columns of U */
            if( k > 1 ) {
                /* Store U(k) and U(k-1) in columns k and k-1 of A */
                magmablas_zlascl_2x2( MagmaUpper, 
                                      k-1, dW(0, kw-1), lddw, dA(0,k-1), ldda, &iinfo, queue );
             }

             /* Copy D(k) to A */
             //magma_zcopymatrix( 2,2, dW(k-1,kw-1), lddw, dA( k-1, k-1 ), ldda, queue );
             *dA( k-1, k-1 ) = *dW(k-1, kw-1);
             *dA( k-1, k   ) = *dW(k-1, kw  );
             *dA( k  , k-1 ) = *dW(k  , kw-1);
             *dA( k  , k   ) = *dW(k  , kw  );

             /* Conjugate W(k) and W(k-1) */
             #if defined(PRECISION_z) || defined(PRECISION_c)
             magmablas_zlacpy_cnjg( k,   dW( 0, kw ),1, dW( 0, kw ),1, queue );
             magmablas_zlacpy_cnjg( k-1, dW( 0, kw-1 ), 1, dW( 0, kw-1 ), 1, queue );
             #endif
           }
          }

          /* Store details of the interchanges in ipiv */
          if( kstep == 1 ) {
            ipiv[ k ] = 1+kp;
          } else {
            ipiv[ k ] = -(1+kp);
            ipiv[ k-1 ] = -(1+kp);
          }
       }
       /* Update the upper triangle of A11 (= A(1:k,1:k)) as
           A11 := A11 - U12*D*U12' = A11 - U12*W'
          computing blocks of NB columns at a time (note that conjg(W) is
          actually stored) */
       kw = nb - (n-k);
       for (magma_int_t j = ( k / nb )*nb; j >= 0; j -= nb ) {
         magma_int_t jb = min( nb, k-j+1 );

         #ifdef SYMMETRIC_UPDATE
         /* Update the upper triangle of the diagonal block */
         for (magma_int_t jj = j; jj < j + jb; jj++) {
            #if defined(PRECISION_z) || defined(PRECISION_c)
            MAGMA_Z_IMAG( *dA(jj, jj) ) = 0.;
            #endif
            lm = jj-j+1;
            ln = n-(k+1); 
            zgemv( lapack_const(MagmaNoTrans), &lm, &ln, 
                   (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( j, k+1 ), &ldda, 
                   (MKL_Complex16 *)dW( jj, kw+1 ), &lddw, (MKL_Complex16 *)&c_one,
                   (MKL_Complex16 *)dA( j, jj ), &ione );
            #if defined(PRECISION_z) || defined(PRECISION_c)
            MAGMA_Z_IMAG(*dA(jj, jj) ) = 0.;
            #endif
         }
         /* Update the rectangular superdiagonal block */
         lk = n-(k+1);
         zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
                 &j, &jb, &lk,
                (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( 0, k+1 ), &ldda, 
                (MKL_Complex16 *)dW( j, kw+1 ), &lddw, (MKL_Complex16 *)&c_one, 
                (MKL_Complex16 *)dA( 0, j ), &ldda );
         #else
         ln = 2*(1+ldda);
         #if defined(PRECISION_z)
         dlaset(lapack_const(MagmaUpperLower), &ione, &jb, &d_zero, &d_zero, 
                ((magmaDouble_ptr)dA(j,j))+1, &ln );
         #elif defined(PRECISION_c)
         slaset(lapack_const(MagmaUpperLower), &ione, &jb, &f_zero, &f_zero,
                ((magmaFloat_ptr)dA(j,j))+1, &ln );
         #endif
         lm = j+jb;
         ln = jb;
         lk = n-(k+1);
         zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
                &lm, &ln, &lk,
                (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( 0, k+1 ),  &ldda,
                (MKL_Complex16 *)dW( j, kw+1 ), &lddw,
                (MKL_Complex16 *)&c_one,  (MKL_Complex16 *)dA( 0, j ),  &ldda );

         ln = 2*(1+ldda); 
         #if defined(PRECISION_z)
         dlaset(lapack_const(MagmaUpperLower), &ione, &jb, &d_zero, &d_zero, 
                ((magmaDouble_ptr)dA(j,j))+1, &ln );
         #elif defined(PRECISION_c)
         slaset(lapack_const(MagmaUpperLower), &ione, &jb, &f_zero, &f_zero,
                ((magmaFloat_ptr)dA(j,j))+1, &ln );
         #endif
         #endif
       }

       /* Put U12 in standard form by partially undoing the interchanges in columns k+1:n */
       for (magma_int_t j = k+1; j < n;)
        {
          magma_int_t jj = j;
          magma_int_t jp = ipiv[ j ];
          if( jp < 0 ) {
            jp = -jp;
            j = j + 1;
          }
          j = j + 1;
          jp = jp - 1;
          if( jp != jj && j < n ){
            ln = n-j;
            zswap( &ln, (MKL_Complex16 *)dA( jp, j ), &ldda, (MKL_Complex16 *)dA( jj, j ), &ldda);
	  }
        }

       /* Set KB to the number of columns factorized */
       *kb = n - (k+1);
 
    } else {
       /* Factorize the leading columns of A using the lower triangle
          of A and working forwards, and compute the matrix W = L21*D
          for use in updating A22 (note that conjg(W) is actually stored)

          K is the main loop index, increasing from 1 in steps of 1 or 2 */

       magma_int_t k;
       for (k = 0; k < min(nb-1,n); k += kstep) {
           /* Copy column K of A to column K of W and update it */
           /* -------------------------------------------------------------- */
           magma_zcopy_mic( n-k, dA( k, k ), 1, dW( k, k ), 1, queue );

           // set imaginary part of diagonal to be zero
           #if defined(PRECISION_z) || defined(PRECISION_c)
           MAGMA_Z_IMAG( *dW( k, k ) ) = 0.; 
           #endif
           /* -------------------------------------------------------------- */

           lm = n-k;
           ln = k;
           zgemv( lapack_const(MagmaNoTrans), &lm, &ln, (MKL_Complex16 *)&c_mone, 
                  (MKL_Complex16 *)dA( k, 0 ), &ldda, (MKL_Complex16 *)dW( k, 0 ), &lddw, 
                  (MKL_Complex16 *)&c_one, (MKL_Complex16 *)dW( k, k ), &ione );
           // re-set imaginary part of diagonal to be zero
           #if defined(PRECISION_z) || defined(PRECISION_c)
           MAGMA_Z_IMAG( *dW( k, k ) ) = 0.;
           #endif

           kstep = 1;

           /* Determine rows and columns to be interchanged and whether
              a 1-by-1 or 2-by-2 pivot block will be used */
           abs_akk = fabs( MAGMA_Z_REAL( *dW( k, k ) ) );

           /* imax is the row-index of the largest off-diagonal element in
              column K, and colmax is its absolute value */
           if( k < n-1 ) {
               // magmablas is one-base
               lm = n-k-1;
               imax = k + izamax( &lm, (MKL_Complex16 *)dW(k+1,k), &ione );
               colmax = MAGMA_Z_ABS1( *dW( imax,k ) );
 
           } else {
               colmax = d_zero;
           }

           if ( max( abs_akk, colmax ) == 0.0 ) {

               /* Column K is zero: set INFO and continue */
               if( info == 0 ) info = k;
               kp = k;

               // make sure the imaginary part of diagonal is zero
               #if defined(PRECISION_z) || defined(PRECISION_c)
               MAGMA_Z_IMAG( *dA( k, k ) ) = 0.; 
               #endif
           } else {
               if ( abs_akk >= alpha*colmax ) {

                   /* no interchange, use 1-by-1 pivot block */

                   kp = k;
               } else {
                   /* Copy column imax to column K+1 of W and update it */
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   magmablas_zlacpy_cnjg( imax-k, dA(imax,k), ldda, dW(k,k+1), 1, queue );
                   #else 
                   magma_zcopy_mic( imax-k, dA( imax, k ), ldda, dW( k, k+1 ), 1, queue );
                   #endif 

                   magma_zcopy_mic( n-imax, dA( imax, imax ), 1, dW( imax, k+1 ), 1, queue );
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   MAGMA_Z_IMAG( *dW( imax, k+1 ) ) = 0.;
                   #endif

                   lm = n-k;
                   ln = k;
                   zgemv( lapack_const(MagmaNoTrans), &lm, &ln, 
                          (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( k, 0 ), &ldda, 
                          (MKL_Complex16 *)dW( imax, 0 ), &lddw, 
                          (MKL_Complex16 *)&c_one, (MKL_Complex16 *)dW( k, k+1 ), &ione );
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   MAGMA_Z_IMAG( *dW( imax, k+1 ) ) = 0.;
                   #endif

                   Zimax = *dW(imax,k+1);

                   /* jmax is the column-index of the largest off-diagonal
                      element in row imax, and rowmax is its absolute value */
 
                   // magmablas is one-base
                   lm = imax-k;
                   jmax = k-1 + izamax( &lm, (MKL_Complex16 *)dW(k, k+1), &ione );

                   rowmax = MAGMA_Z_ABS1( *dW(jmax,k+1) );
                   if( imax < n-1 ) {
                       // magmablas is one-base
                       lm = (n-1)-imax;
                       jmax = imax + izamax( &lm, (MKL_Complex16 *)dW(imax+1,k+1), &ione );
                       rowmax = max( rowmax, MAGMA_Z_ABS1( *dW(jmax,k+1) ) );
                   }

                   if( abs_akk >= alpha*colmax*( colmax / rowmax ) ) {

                       /* no interchange, use 1-by-1 pivot block */
                       kp = k;
                   } else if( fabs( MAGMA_Z_REAL( Zimax ) ) >= alpha*rowmax ) {

                       /* interchange rows and columns K and imax, use 1-by-1
                          pivot block */
                       kp = imax;

                       /* copy column K+1 of W to column K */
                       magma_zcopy_mic( n-k, dW( k, k+1 ), 1, dW( k, k ), 1, queue );
                   } else {

                       /* interchange rows and columns K+1 and imax, use 2-by-2
                          pivot block */
                       kp = imax;
                       kstep = 2;
                   }
               }

               kk = k + kstep - 1;

               /* Updated column kp is already stored in column kk of W */
               if( kp != kk ) {

                   /* Copy non-updated column kk to column kp */
                   /* ------------------------------------------------------------------ */
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   magmablas_zlacpy_cnjg( kp-kk, dA( kk, kk ), 1, dA( kp, kk ), ldda, queue );
                   #else
                   magma_zcopy_mic( kp-kk, dA( kk, kk ), 1, dA( kp, kk ), ldda, queue );
                   #endif
                   if ( kp < n ) {
                   magma_zcopy_mic( n-kp, dA( kp, kk), 1, dA( kp, kp ), 1, queue );
                   }
                   /* ------------------------------------------------------------------ */

                   /* Interchange rows kk and kp in first kk columns of A and W */
                   ln = kk+1;
                   zswap( &ln, (MKL_Complex16 *)dA( kk, 0 ), &ldda, (MKL_Complex16 *)dA( kp, 0 ), &ldda );
                   zswap( &ln, (MKL_Complex16 *)dW( kk, 0 ), &lddw, (MKL_Complex16 *)dW( kp, 0 ), &lddw );
               }

               if ( kstep == 1 ) {
 
                   /* 1-by-1 pivot block D(k): column k of W now holds

                      W(k) = L(k)*D(k)

                      where L(k) is the k-th column of L
 
                      Store L(k) in column k of A */
                   magma_zcopy_mic( n-k, dW( k, k ), 1, dA( k, k ), 1, queue );

                   if ( k < n-1 ) {
                       R1 = d_one / MAGMA_Z_REAL( *dA(k,k) );
                       magma_zdscal_mic((n-1)-k, R1, dA( k+1,k ), 1, queue);

                       /* Conjugate W(k) */
                       #if defined(PRECISION_z) || defined(PRECISION_c)
                       magmablas_zlacpy_cnjg( (n-1)-k, dW( k+1,k ),1, dW( k+1,k ),1, queue );
                       #endif
                   }
               } else {

                   /* 2-by-2 pivot block D(k): columns k and k+1 of W now hold

                   ( W(k) W(k+1) ) = ( L(k) L(k+1) )*D(k)

                   where L(k) and L(k+1) are the k-th and (k+1)-th columns
                   of L */
                   magmablas_zlascl_2x2( MagmaLower,
                                         n-(k+2), dW(k,k), lddw, dA(k+2,k), ldda, &iinfo, 
                                         queue );

                   /* Copy D(k) to A */
                   //magma_zcopymatrix( 2,2, dW( k, k ), lddw, dA( k, k ), ldda, queue );
                   *dA( k  , k  ) = *dW( k  , k  );
		   *dA( k  , k+1) = *dW( k  , k+1);
		   *dA( k+1, k  ) = *dW( k+1, k  );
		   *dA( k+1, k+1) = *dW( k+1, k+1);

                   /* Conjugate W(k) and W(k+1) */
                   #if defined(PRECISION_z) || defined(PRECISION_c)
                   magmablas_zlacpy_cnjg( (n-1)-k,   dW( k+1,k ),1, dW( k+1,k ),1, queue );
                   magmablas_zlacpy_cnjg( (n-1)-k-1, dW( k+2,k+1),1, dW( k+2,k+1 ),1, queue );
                   #endif
               }
           }

           /* Store details of the interchanges in ipiv */
           if ( kstep == 1 ) {
               ipiv[k] = kp+1;
           } else {
               ipiv[k] = -kp-1;
               ipiv[k+1] = -kp-1;
           }
       } 

       /* Update the lower triangle of A22 (= A(k:n,k:n)) as

          A22 := A22 - L21*D*L21' = A22 - L21*W'

          computing blocks of NB columns at a time (note that conjg(W) is
          actually stored) */
       for( magma_int_t j = k; j < n; j += nb ) {
           magma_int_t jb = min( nb, n-j );

           /* Update the lower triangle of the diagonal block */

           #ifdef SYMMETRIC_UPDATE
           for (magma_int_t jj = j; jj < j + jb; jj++) {
               magma_int_t jnb = j + jb - jj;

               /* -------------------------------------------------------- */
               zgemv( lapack_const(MagmaNoTrans), &jnb, &k, 
                      (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( jj, 0 ), &ldda, 
                      (MKL_Complex16 *)dW( jj, 0 ), &lddw, (MKL_Complex16 *)c_one, 
                      (MKL_Complex16 *)dA( jj, jj ), &ione );
               /* -------------------------------------------------------- */
           }

           /* Update the rectangular subdiagonal block */

           if( j+jb < n ) {
               magma_int_t nk = n - (j+jb);

               /* -------------------------------------------- */
               zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
                      &nk, &jb, &k, 
                      (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( j+jb, 0 ), &ldda, 
                      (MKL_Complex16 *)dW( j, 0 ),    &lddw,
                      (MKL_Complex16 *)&c_one, (MKL_Complex16 *)dA( j+jb, j ), &ldda );
               /* ------------------------------------------- */
           }
           #else

           ln = 2*(1+ldda);
           #if defined(PRECISION_z)
           dlaset(lapack_const(MagmaUpperLower), &ione, &jb, &d_zero, &d_zero,  
                  ((magmaDouble_ptr)dA(j,j))+1, &ln );
           #elif defined(PRECISION_c)
           slaset(lapack_const(MagmaUpperLower), &ione, &jb, &f_zero, &f_zero,
                  ((magmaFloat_ptr)dA(j,j))+1, &ln );  
           #endif
           lm = n-j;
           zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), &lm, &jb, &k, 
                  (MKL_Complex16 *)&c_mone, (MKL_Complex16 *)dA( j, 0 ), &ldda, 
                  (MKL_Complex16 *)dW( j, 0 ), &lddw,
                  (MKL_Complex16 *)&c_one, (MKL_Complex16 *)dA( j, j ), &ldda );

           ln =2*(1+ldda); 
           #if defined(PRECISION_z)
           dlaset(lapack_const(MagmaUpperLower), &ione, &jb, &d_zero, &d_zero,  
                  ((magmaDouble_ptr)dA(j,j))+1, &ln );
           #elif defined(PRECISION_c)
           slaset(lapack_const(MagmaUpperLower), &ione, &jb, &f_zero, &f_zero,
                  ((magmaFloat_ptr)dA(j,j))+1, &ln );
           #endif
           #endif
       }

       /* Put L21 in standard form by partially undoing the interchanges
          in columns 1:k-1 */
       for (magma_int_t j = k; j > 0;) {
           magma_int_t jj = j;
           magma_int_t jp = ipiv[j-1];
           if( jp < 0 ) {
               jp = -jp;
               j--;
           }
           j--;
           if ( jp != jj && j >= 1 ) {
	     zswap( &j, (MKL_Complex16 *)dA( jp-1,0 ), &ldda, (MKL_Complex16 *)dA( jj-1,0 ), &ldda );
           }
       }

       /* Set KB to the number of columns factorized */
       *kb = k;
    }

    /* End of ZLAHEF */
}

