/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @generated from magma_z_blaswrapper.cpp normal z -> s, Sun Jul 12 17:46:30 2015
       @author Hartwig Anzt

*/

//#include <cusparse_v2.h>

#include "common_magma.h"
#include "magmablas.h"
#include "magmasparse_types.h"
#include "magmasparse.h"




/**
    Purpose
    -------

    For a given input matrix A and vectors x, y and scalars alpha, beta
    the wrapper determines the suitable SpMV computing
              y = alpha * A * x + beta * y.  
    Arguments
    ---------

    @param[in]
    alpha       float
                scalar alpha

    @param[in]
    A           magma_s_sparse_matrix
                sparse matrix A    

    @param[in]
    x           magma_s_vector
                input vector x  
                
    @param[in]
    beta        float
                scalar beta
    @param[out]
    y           magma_s_vector
                output vector y      
    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_s
    ********************************************************************/

extern "C" magma_int_t
magma_s_spmv(
    float alpha, 
    magma_s_sparse_matrix A, 
    magma_s_vector x, 
    float beta, 
    magma_s_vector y,
    magma_queue_t queue )
{
    // set queue for old dense routines
    //magma_queue_t orig_queue;
    //magmablasGetKernelStream( &orig_queue );

    if ( A.memory_location != x.memory_location || 
                            x.memory_location != y.memory_location ) {
        printf("error: linear algebra objects are not located in same memory!\n");
        printf("memory locations are: %d   %d   %d\n", 
                        A.memory_location, x.memory_location, y.memory_location );
        //magmablasSetKernelStream( orig_queue );
        return MAGMA_ERR_INVALID_PTR;
    }

    // DEV case
    if ( A.memory_location == Magma_DEV ) {
        if ( A.num_cols == x.num_rows && x.num_cols == 1 ) {

             if ( A.storage_type == Magma_CSR 
                            || A.storage_type == Magma_CSRL 
                            || A.storage_type == Magma_CSRU ) {
                    
                magma_scsrgemv( MagmaNoTrans, 
                                A.num_rows, A.dval, A.drow, A.dcol,
                                x.dval, y.dval, queue );
             }
             else {
                 printf("error: format not supported.\n");
                 // magmablasSetKernelStream( orig_queue );
                 return MAGMA_ERR_NOT_SUPPORTED;
             }
        }
        else if ( A.num_cols < x.num_rows || x.num_cols > 1 ) {
            magma_int_t num_vecs = x.num_rows / A.num_cols * x.num_cols;
            if ( A.storage_type == Magma_CSR ) {

                
             }
             else {
                 printf("error: format not supported.\n");
                 // magmablasSetKernelStream( orig_queue );
                 return MAGMA_ERR_NOT_SUPPORTED;
             }
        }
         
         
    }
    // CPU case missing!     
    else {
        printf("error: CPU not yet supported.\n");
        // magmablasSetKernelStream( orig_queue );
        return MAGMA_ERR_NOT_SUPPORTED;
    }
    // magmablasSetKernelStream( orig_queue );
    return MAGMA_SUCCESS;
}

/**
    Purpose
    -------

    For a given input matrix A and B and scalar alpha,
    the wrapper determines the suitable SpMV computing
              C = alpha * A * B.  
    Arguments
    ---------

    @param[in]
    alpha       float
                scalar alpha

    @param[in]
    A           magma_s_sparse_matrix
                sparse matrix A    
                
    @param[in]
    B           magma_s_sparse_matrix
                sparse matrix C    
                
    @param[out]
    C           magma_s_sparse_matrix *
                outpur sparse matrix C    

    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_s
    ********************************************************************/

extern "C" magma_int_t
magma_s_spmm(
    float alpha, 
    magma_s_sparse_matrix A,
    magma_s_sparse_matrix B,
    magma_s_sparse_matrix *C,
    magma_queue_t queue )
{
    // set queue for old dense routines
    //magma_queue_t orig_queue;
    //magmablasGetKernelStream( &orig_queue );

    if ( A.memory_location != B.memory_location ) {
        printf("error: linear algebra objects are not located in same memory!\n");
        printf("memory locations are: %d   %d\n", 
                        A.memory_location, B.memory_location );
        //magmablasSetKernelStream( orig_queue );
        return MAGMA_ERR_INVALID_PTR;
    }

    // DEV case
    if ( A.memory_location == Magma_DEV ) {
        if ( A.num_cols == B.num_rows ) {

             if ( A.storage_type == Magma_CSR 
                            || A.storage_type == Magma_CSRL 
                            || A.storage_type == Magma_CSRU ) {
                    
                magma_int_t request = 1;
                magma_int_t sort = 13;  // no sort at all
                magma_int_t info = 0;
                magma_int_t nzmax = 0;
                magma_int_t stat_dev = 0;
                
                magma_s_mtransfer( A, C, Magma_DEV, Magma_DEV, queue );
                
                magma_free( C->dcol );
                C->dcol = NULL;
                magma_free( C->dval );
                C->dval = NULL;
               
                
                // first generate nnz info
                magma_scsrgespmm(
                    MagmaNoTrans,
                    request,
                    sort,
                    A.num_rows,
                    A.num_cols,
                    B.num_cols,
                    alpha,
                    A.dval,
                    A.drow,
                    A.dcol,
                    B.dval,
                    B.drow,
                    B.dcol,
                    C->dval,
                    C->drow,
                    C->dcol,
                    nzmax,
                    info,
                    queue );
                                // get nnz info
                magma_index_t *nnz;                
                magma_index_malloc_pinned( &nnz, 1 );
                magma_index_getvector( 1, C->drow+C->num_rows, 1, nnz, 1 , queue );
                C->nnz = (magma_int_t) nnz[0]-1 ;
                
                magma_free_pinned( nnz );
                
                // allocate space
                stat_dev += magma_smalloc( &C->dval, C->nnz ); 
                if( stat_dev != 0 ){ magma_s_mfree( C , queue ); printf("ola\n\n\n" );
                    return MAGMA_ERR_DEVICE_ALLOC;}
                stat_dev += magma_index_malloc( &C->dcol, C->nnz ); 
                if( stat_dev != 0 ){ magma_s_mfree( C , queue ); 
                    return MAGMA_ERR_DEVICE_ALLOC;}

                // now run multiplication
                request = 0;
                magma_scsrgespmm(
                    MagmaNoTrans,
                    request,
                    sort,
                    A.num_rows,
                    A.num_cols,
                    B.num_cols,
                    alpha,
                    A.dval,
                    A.drow,
                    A.dcol,
                    B.dval,
                    B.drow,
                    B.dcol,
                    C->dval,
                    C->drow,
                    C->dcol,
                    nzmax,
                    info,
                    queue );
             }
             else {
                 printf("error: format not supported.\n");
                 // magmablasSetKernelStream( orig_queue );
                 return MAGMA_ERR_NOT_SUPPORTED;
             }
        }
         
    }
    // CPU case missing!     
    else {
        printf("error: CPU not yet supported.\n");
        // magmablasSetKernelStream( orig_queue );
        return MAGMA_ERR_NOT_SUPPORTED;
    }
    // magmablasSetKernelStream( orig_queue );
    return MAGMA_SUCCESS;
}



