/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @generated from magma_zmtranspose.cpp normal z -> s, Sun Jul 12 17:46:30 2015
       @author Hartwig Anzt

*/
#include "magma_lapack.h"
#include "common_magma.h"
#include "magmasparse.h"

#include <assert.h>

/**
    Purpose
    -------

    Transposes a matrix stored in CSR format.


    Arguments
    ---------

    @param[in]
    n_rows      magma_int_t
                number of rows in input matrix

    @param[in]
    n_cols      magma_int_t
                number of columns in input matrix

    @param[in]
    nnz         magma_int_t
                number of nonzeros in input matrix

    @param[in]
    val         float*
                value array of input matrix 

    @param[in]
    row         magma_index_t*
                row pointer of input matrix

    @param[in]
    col         magma_index_t*
                column indices of input matrix 

    @param[in]
    new_n_rows  magma_index_t*
                number of rows in transposed matrix

    @param[in]
    new_n_cols  magma_index_t*
                number of columns in transposed matrix

    @param[in]
    new_nnz     magma_index_t*
                number of nonzeros in transposed matrix

    @param[in]
    new_val     float**
                value array of transposed matrix 

    @param[in]
    new_row     magma_index_t**
                row pointer of transposed matrix

    @param[in]
    new_col     magma_index_t**
                column indices of transposed matrix

    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_saux
    ********************************************************************/

extern "C" magma_int_t 
s_transpose_csr(    
    magma_int_t n_rows, 
    magma_int_t n_cols, 
    magma_int_t nnz, 
    float *val, 
    magma_index_t *row, 
    magma_index_t *col, 
    magma_int_t *new_n_rows, 
    magma_int_t *new_n_cols, 
    magma_int_t *new_nnz, 
    float **new_val, 
    magma_index_t **new_row, 
    magma_index_t **new_col,
    magma_queue_t queue )
{



  nnz = row[n_rows];
  *new_n_rows = n_cols;
  *new_n_cols = n_rows;
  *new_nnz = nnz;

  float ** valtemp;
  magma_index_t ** coltemp;
  valtemp = (float**)malloc((n_rows)*sizeof(float*));
  coltemp =(magma_index_t**)malloc((n_rows)*sizeof(magma_index_t*));

  // temporary 2-dimensional arrays valtemp/coltemp 
  // where val[i] is the array with the values of the i-th column of the matrix
  magma_index_t *nnztemp;
  magma_index_malloc_pinned( &nnztemp, n_rows );
  
  for( magma_int_t i=0; i<n_rows; i++ )
    nnztemp[i]=0;
  for( magma_int_t i=0; i<nnz; i++ )
    nnztemp[col[i]]++;    

  for( magma_int_t i=0; i<n_rows; i++ ){
    valtemp[i] = 
        (float*)malloc((nnztemp[i])*sizeof(float));
    coltemp[i] = (magma_index_t*)malloc(nnztemp[i]*sizeof(magma_index_t));
  }

  for( magma_int_t i=0; i<n_rows; i++ )
    nnztemp[i]=0;

  for( magma_int_t j=0; j<n_rows; j++ ){
    for( magma_int_t i=row[j]; i<row[j+1]; i++ ){
      valtemp[col[i]][nnztemp[col[i]]]=val[i];printf("%f\n", val[i]);
      coltemp[col[i]][nnztemp[col[i]]]=j;
      nnztemp[col[i]]++;    
    }
  }

  //csr structure for transposed matrix
  *new_val = new float[nnz];
  *new_row = new magma_index_t[n_rows+1];
  *new_col = new magma_index_t[nnz];

  //fill the transposed csr structure
  magma_int_t nnztmp=0;
  (*new_row)[0]=0;
  for( magma_int_t j=0; j<n_rows; j++ ){
    for( magma_int_t i=0; i<nnztemp[j]; i++ ){
      (*new_val)[nnztmp]=valtemp[j][i];
      (*new_col)[nnztmp]=coltemp[j][i];
      nnztmp++; printf("%f\n", coltemp[j][i]);
    }
    (*new_row)[j+1]=nnztmp;
  }

//usually the temporary memory should be freed afterwards
//however, it does not work
/*
  for( magma_int_t j=0; j<n_rows; j++ ){
    free(valtemp[j]);
    free(coltemp[j]);
  }
  free(valtemp);free(coltemp);
      printf("check9\n");
    fflush(stdout);
*/

    magma_free_pinned( nnztemp );
    
    return MAGMA_SUCCESS;
}


/**
    Purpose
    -------

    Interface to cuSPARSE transpose.

    Arguments
    ---------

    @param[in]
    A           magma_s_sparse_matrix
                input matrix (CSR)

    @param[out]
    B           magma_s_sparse_matrix*
                output matrix (CSR)
    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_saux
    ********************************************************************/
    
    
extern "C" magma_int_t
magma_s_mtranspose(
    magma_s_sparse_matrix A, magma_s_sparse_matrix *B,
    magma_queue_t queue )
{
    
    magma_s_sparse_matrix T;
    magma_s_mconvert( A, &T, Magma_CSR, Magma_CSRCOO, queue );
    magma_s_mconvert( A, B, Magma_CSR, Magma_CSR, queue );

    for( magma_int_t i=0; i<B->num_rows+1; i++ ){    
        B->row[i]=0;
    }
    magma_int_t nnz=0;
    for( magma_int_t i=0; i<T.num_rows; i++ ){    
        for( magma_int_t j=0; j<T.nnz; j++ ){
            if( T.col[j] == i ){
                 B->col[nnz] = T.rowidx[j];
                 B->val[nnz] = T.val[j];
                 B->row[T.col[j]+1]++;
                 nnz++;
            }
        }
    }
    B->row[0]=0;
    for( magma_int_t i=1; i<B->num_rows+1; i++ ){  
        B->row[i]=B->row[i]+B->row[i-1];
    }
    
    magma_s_mfree( &T, queue );
    
    // magma_s_cucsrtranspose( A, B, queue );
    return MAGMA_SUCCESS;

}


/**
    Purpose
    -------

    Helper function to transpose CSR matrix. 
    Using the CUSPARSE CSR2CSC function.


    Arguments
    ---------

    @param[in]
    A           magma_s_sparse_matrix
                input matrix (CSR)

    @param[out]
    B           magma_s_sparse_matrix*
                output matrix (CSR)
    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_saux
    ********************************************************************/

extern "C" magma_int_t
magma_s_cucsrtranspose(
    magma_s_sparse_matrix A, 
    magma_s_sparse_matrix *B,
    magma_queue_t queue )
{
    // for symmetric matrices: convert to csc using cusparse

    if( A.storage_type == Magma_CSR && A.memory_location == Magma_DEV ) {
                  
         magma_s_sparse_matrix C;
         magma_s_mtransfer( A, &C, Magma_DEV, Magma_DEV, queue );
         
         // here should go the MKL transpose routine!
         // still missing!
         
         printf("error: not yet supported! \n");

        return MAGMA_SUCCESS;
        
    }else if( A.storage_type == Magma_CSR && A.memory_location == Magma_CPU ){
               
        magma_s_sparse_matrix A_d, B_d;

        magma_s_mtransfer( A, &A_d, A.memory_location, Magma_DEV, queue );
        magma_s_cucsrtranspose( A_d, &B_d, queue );
        magma_s_mtransfer( B_d, B, Magma_DEV, A.memory_location, queue );
        
        magma_s_mfree( &A_d, queue );
        magma_s_mfree( &B_d, queue );
        
        return MAGMA_SUCCESS;
                
    }else {

        magma_s_sparse_matrix ACSR, BCSR;
        
        magma_s_mconvert( A, &ACSR, A.storage_type, Magma_CSR, queue );
        magma_s_cucsrtranspose( ACSR, &BCSR, queue );
        magma_s_mconvert( BCSR, B, Magma_CSR, A.storage_type, queue );
       
        magma_s_mfree( &ACSR, queue );
        magma_s_mfree( &BCSR, queue );

        return MAGMA_SUCCESS;
    }
}



