/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @precisions normal z -> c d s
       @author Hartwig Anzt

*/
#include "common_magma.h"
#include "magmasparse.h"




/**
    Purpose
    -------

    ALlows the user to choose a solver.

    Arguments
    ---------

    @param[in]
    A           magma_z_sparse_matrix
                sparse matrix A    

    @param[in]
    b           magma_z_vector
                input vector b     

    @param[in]
    x           magma_z_vector*
                output vector x        

    @param[in]
    zopts     magma_zopts
              options for solver and preconditioner
    @param[in]
    queue       magma_queue_t
                Queue to execute in.

    @ingroup magmasparse_zaux
    ********************************************************************/

extern "C" magma_int_t
magma_z_solver(
    magma_z_sparse_matrix A, magma_z_vector b, 
    magma_z_vector *x, magma_zopts *zopts,
    magma_queue_t queue )
{
    // preconditioner
        if ( zopts->solver_par.solver != Magma_ITERREF ) {
            magma_z_precondsetup( A, b, &zopts->precond_par, queue );
        }
        switch( zopts->solver_par.solver ) {
            case  Magma_CG:
                magma_zcg( A, b, x, &zopts->solver_par, queue );break;
            case  Magma_CG_mic:
                {
                    magmaDoubleComplex_ptr dwork;
                    magma_zmalloc(&dwork, 3*A.num_rows);
                    magma_zcg_mic(A.num_rows,
                                  A.dval, A.drow, A.dcol,
                                  b.dval, x->dval, &zopts->solver_par, dwork);
                    magma_free(dwork);
                    break;
                }
            case  Magma_CGMERGE:
                printf("error: not supported\n");break;//  magma_zcg_merge( A, b, x, &zopts->solver_par, queue );break;
            case  Magma_PCG:
                    magma_zpcg( A, b, x, &zopts->solver_par, &zopts->precond_par, queue );break;
            case  Magma_BICGSTAB:
                      magma_zbicgstab( A, b, x, &zopts->solver_par, queue );break;
            case  Magma_BICGSTAB_mic:
                {
                    magmaDoubleComplex_ptr dwork;
                    magma_zmalloc(&dwork, 6*A.num_rows);
                    magma_zbicgstab_mic(A.num_rows,
                                        A.dval, A.drow, A.dcol,
                                        b.dval, x->dval, &zopts->solver_par, dwork);
                    magma_free(dwork);
                    break;
                }
            case  Magma_BICGSTABMERGE: 
                    printf("error: not supported\n");break;//  magma_zbicgstab_merge( A, b, x, &zopts->solver_par, queue );break;
            case  Magma_PBICGSTAB: 
                  break;//  magma_zpbicgstab( A, b, x, &zopts->solver_par, &zopts->precond_par, queue );break;
            case  Magma_GMRES: 
                      magma_zfgmres( A, b, x, &zopts->solver_par, &zopts->precond_par, queue );break;
            case  Magma_GMRES_mic:
                {
                    magmaDoubleComplex_ptr dwork;
                    magma_int_t r = zopts->solver_par.restart;
                    magma_zmalloc(&dwork, (r+3)*A.num_rows + (r+1)*(5+2*r));
                    magma_zgmres_mic(A.num_rows,
                                     A.dval, A.drow, A.dcol,
                                     b.dval, x->dval, &zopts->solver_par, dwork);
                    magma_free(dwork);
                    break;
                }
            case  Magma_PGMRES: 
                      magma_zfgmres( A, b, x, &zopts->solver_par, &zopts->precond_par, queue );break;
            case  Magma_LOBPCG: 
                  break;//  magma_zlobpcg( A, &zopts->solver_par, queue );break;
            case  Magma_ITERREF:
                    printf("error: not supported\n");break;//  magma_ziterref( A, b, x, &zopts->solver_par, &zopts->precond_par, queue );break;
            case  Magma_JACOBI: 
                    magma_zjacobi( A, b, x, &zopts->solver_par, queue );break;
            case  Magma_BAITER: 
                    printf("error: not supported\n");break;//  magma_zbaiter( A, b, x, &zopts->solver_par, queue );break;
            
        }
    return MAGMA_SUCCESS;
}


