/*
   -- micMAGMA (version 1.4.0) --
   Univ. of Tennessee, Knoxville
   Univ. of California, Berkeley
   Univ. of Colorado, Denver
   April 2012

   @precisions normal z -> s d c
*/

#include <stdio.h>
#include "common_magma.h"

magma_err_t
magma_zgetrf_mic(magma_int_t m, magma_int_t n, 
		magmaDoubleComplex_ptr dA, size_t dA_offset, magma_int_t ldda,
		magma_int_t *ipiv, magma_int_t *info,
		magma_queue_t queue )
{
/*  -- micMAGMA (version 1.4.0) --
    Univ. of Tennessee, Knoxville
    Univ. of California, Berkeley
    Univ. of Colorado, Denver
    April 2012

    Purpose
    =======
    ZGETRF computes an LU factorization of a general M-by-N matrix A
    using partial pivoting with row interchanges.
    
    The factorization has the form
    A = P * L * U
    where P is a permutation matrix, L is lower triangular with unit
    diagonal elements (lower trapezoidal if m > n), and U is upper
    triangular (upper trapezoidal if m < n).

    This is the right-looking Level 3 BLAS version of the algorithm.
    
    Arguments
    =========
    
    M       (input) INTEGER
            The number of rows of the matrix A.  M >= 0.
    
    N       (input) INTEGER
            The number of columns of the matrix A.  N >= 0.

    A       (input/output) COMPLEX_16 array on the GPU, dimension (LDDA,N).
            On entry, the M-by-N matrix to be factored.
	    On exit, the factors L and U from the factorization
	    A = P*L*U; the unit diagonal elements of L are not stored.

    LDDA   (input) INTEGER
           The leading dimension of the array A.  LDDA >= max(1,M).

    IPIV   (output) INTEGER array, dimension (min(M,N))
           The pivot indices; for 1 <= i <= min(M,N), row i of the
           matrix was interchanged with row IPIV(i).
	   
    INFO   (output) INTEGER
           = 0:  successful exit
	   < 0:  if INFO = -i, the i-th argument had an illegal value
	         if INFO = -7, internal GPU memory allocation failed.
	   > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
	         has been completed, but the factor U is exactly
		 singular, and division by zero will occur if it is used
		 to solve a system of equations.
=====================================================================    */

#define inA(i,j) (dA + dA_offset + (j) * nb * ldda + (i) * nb)
#define tmppanel(i_) (tmppanel + (i_) * nb)
#define tmpprevpanel(i_) (tmpprevpanel + (i_) * nb)

	magmaDoubleComplex c_one     = MAGMA_Z_MAKE(  1.0, 0.0 );
	magmaDoubleComplex c_neg_one = MAGMA_Z_MAKE( -1.0, 0.0 );

	magma_int_t lwork, iinfo, nb;
	magma_int_t mindim;
	magma_int_t i, rows, s, lddwork;

	magma_int_t *d_ipiv;

	magmaDoubleComplex *work, *tmppanel0, *tmppanel1, *tmppanel, *tmpprevpanel;

	magma_err_t err;

	*info = 0;
	int jj;

	if (m < 0)
		*info = -1;
	else if (n < 0)
		*info = -2;
	else if (ldda < max(1,m))
		*info = -4;

	if (*info != 0) {
		magma_xerbla( __func__, -(*info) );
		return *info;
	}

	if (m == 0 || n == 0)
		return MAGMA_SUCCESS;

	mindim = min(m, n);
	//if(mindim >= 24000)
	//	nb = 720;
	//else 
		nb = 480;
	//nb     = 120;//720;//480;//720;//240;//480;//128;//256;//magma_get_zgetrf_nb(m);
	s      = mindim / nb;

	magma_event_t event[2] = {NULL, NULL}; 

	if (nb < 1 || nb >= min(m,n)) 
	{
		// use CPU code
		err = magma_zmalloc_pinned( &work, m*n );
		if ( err != MAGMA_SUCCESS ) {
			*info = MAGMA_ERR_HOST_ALLOC;
			return *info;
		}

		chk( magma_zgetmatrix( m, n, dA, dA_offset, ldda, work, 0, m, queue ));
		lapackf77_zgetrf(&m, &n, work, &m, ipiv, info);
		chk( magma_zsetmatrix( m, n, work, 0, m, dA, dA_offset, ldda, queue ));

		magma_free_pinned(work);
	}
	else 
	{
		//#define ENABLE_TIMER
		double ttrsmgemm, tmtc, tipiv, tgetrf, tswap, tctm, tcnp, tmnp;

		// use hybrid blocked code
		lddwork = ldda;

		//memory allocate for d_ipiv
		if ( MAGMA_SUCCESS != magma_imalloc( &d_ipiv, ldda )) {
			*info = MAGMA_ERR_DEVICE_ALLOC;
			return *info;
		}

		if ( MAGMA_SUCCESS != magma_zmalloc_pinned( &work, 2 * ldda * nb ) ) {
			*info = MAGMA_ERR_HOST_ALLOC;
			return *info;
		}

		#if defined (ENABLE_TIMER)
		printf("\n\n %10s %10s %10s  %10s %10s %10s %10s  %10s  %10s %10s  %10s\n",
			"i", "row", "panel", "tipiv", "CPU", "trsmgemm", "mtswap", "CPU_np", "MIC_np", "ctm P", "mtc");
		printf("     ====================================================================================================\n");
		#endif


		tmppanel0 = work;
		tmppanel1 = work + nb * ldda;
		//next_panel = work + nb * ldda;

		for(i = 0; i < s; i++)
		{
			magma_kernel_sync();
			tmppanel = (i % 2) == 0 ? tmppanel0 : tmppanel1;

			#if defined (ENABLE_TIMER)
			ttrsmgemm = tmtc = tgetrf = tswap = tctm = tmnp = tcnp = tipiv = 0.0;
			#endif

			if ( i>0 ){

				#if defined (ENABLE_TIMER)
				ttrsmgemm = get_time();
				#endif

				magma_ztrsmgemm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit, 
						nb, n - (i+1) * nb, 
						c_one, inA(i - 1, i - 1), 0, ldda, 
						inA(i - 1, i+1), 0, ldda,
						MagmaNoTrans, MagmaNoTrans, 
						m - i * nb, n - (i+1) * nb, nb, 
						c_neg_one, inA(i, i - 1), 0, ldda, 
						inA(i - 1,  i + 1), 0, ldda, 
						c_one,     inA(i,  i+1), 0, ldda, queue );
				
				#if defined (ENABLE_TIMER)
				magma_kernel_sync();
				ttrsmgemm = get_time() - ttrsmgemm;
				#endif

				#if defined (ENABLE_TIMER)
				tcnp = get_time();
				#endif

				//donwload look ahead panel
				tmpprevpanel = ((i - 1) % 2) == 0 ? tmppanel0 : tmppanel1;
				magma_zgetmatrix_async(m - (i - 1) * nb, nb, inA(i - 1, i), 0, ldda, tmppanel(i - 1), 0, lddwork, queue, &event[0]);
				magma_event_sync(event[0]);
				
				//update of look ahead panel on cpu
				ztrsm_( "L", "L", "N", "U", 
						&nb, &nb, 
						&c_one, tmpprevpanel(i - 1), &lddwork, 
						tmppanel(i - 1), &lddwork);

				rows = m - i * nb;
				zgemm_( "N", "N", 
						&rows, &nb, &nb, 
						&c_neg_one, tmpprevpanel(i - 1) + nb, &lddwork,
						tmppanel(i - 1), &lddwork, 
						&c_one,     tmppanel(i - 1) + nb, &lddwork );

				//send back look ahead panel to mic
				//magma_zsetmatrix_async(m - (i - 1) * nb, nb, next_panel, 0, lddwork, inA(i - 1, i), 0, ldda, queue, &event[1]);
				magma_zsetmatrix_async(nb, nb, tmppanel(i - 1), 0, lddwork, inA(i - 1, i), 0, ldda, queue, &event[1]);
				magma_event_sync(event[1]);
				
				#if defined (ENABLE_TIMER)
				tcnp = get_time() - tcnp;
				#endif
	
			}

			if(i == 0) {
				#if defined (ENABLE_TIMER)
				tmtc = get_time();
				#endif

				// download i-th panel
				magma_zgetmatrix_async(m - i * nb, nb, inA(i, i), 0, ldda, tmppanel(i), 0, lddwork, queue, &event[0]);
				magma_event_sync(event[0]);

				#if defined (ENABLE_TIMER)
				tmtc = get_time() - tmtc;
				#endif

			}


			#if defined (ENABLE_TIMER)
			tgetrf = get_time();
			#endif

			// do the cpu part
			rows = m - i * nb;
			lapackf77_zgetrf( &rows, &nb, tmppanel(i), &lddwork, ipiv + i * nb, &iinfo);
			if ( (*info == 0) && (iinfo > 0) )
				*info = iinfo + i*nb;

			#if defined (ENABLE_TIMER)
			tgetrf = get_time() - tgetrf;
			#endif

			#if defined (ENABLE_TIMER)
			tipiv = get_time();
			#endif
			//transfer ipiv 
 			magma_zsetintarray_async(mindim, ipiv, 0, d_ipiv, 0, queue, &event[1]);
			magma_event_sync(event[1]);

			#if defined (ENABLE_TIMER)
			tipiv = get_time() - tipiv;
			#endif

			#if defined (ENABLE_TIMER)
			tswap = get_time();
			#endif
			magma_zlaswp(n, inA(i, 0), ldda, 1, nb, d_ipiv + i * nb, 1, queue);

			for (jj = 0; jj < nb; jj++) {
				*(ipiv + i*nb + jj) += i*nb;
			}

			magma_kernel_sync();

			#if defined (ENABLE_TIMER)
			magma_kernel_sync();
			tswap = get_time() - tswap;
			#endif

			#if defined (ENABLE_TIMER)
			tctm = get_time();
			#endif

			// upload i-th panel
			magma_zsetmatrix_async(m - i * nb, nb, tmppanel(i), 0, lddwork, inA(i, i), 0, ldda, queue, &event[1]);
			magma_event_sync(event[1]);

			#if defined (ENABLE_TIMER)
			tctm = get_time() - tctm;
			#endif

			// non-parallel computations -- last block
			if(s <= (i + 1)) {
				magma_ztrsmgemm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit, 
						nb, n - s * nb, 
						c_one, inA(i, i), 0, ldda, 
						inA(i, i+1), 0, ldda, 
						MagmaNoTrans, MagmaNoTrans, 
						m - (i + 1) * nb, n - (i + 1) * nb, nb,
						c_neg_one, inA(i + 1,   i), 0, ldda,
						inA(i, i+1  ), 0, ldda, 
						c_one,     inA(i+1, i+1), 0, ldda, queue );
		
			}

			#if defined (ENABLE_TIMER)
			printf("%10d %10d %10.3lf  %10.3lf  %10.3lf %10.3lf %10.3lf  %10.3lf %10.3lf  %10.3lf %10.3lf\n",
					i, rows, tgetrf * 1000, tipiv*1000, (tgetrf+tipiv+tcnp)*1000, ttrsmgemm * 1000, 
					tswap * 1000, tcnp * 1000, tmnp * 1000, tctm * 1000, tmtc * 1000);
			#endif

		}

		magma_int_t nb0 = min(m - s * nb, n - s * nb);
		rows = m - s * nb;
		
		if(nb0 > 0) {
			magma_zgetmatrix(rows, nb0, inA(s, s), 0, ldda, tmppanel0, 0, lddwork, queue);

			// do the cpu part
			lapackf77_zgetrf( &rows, &nb0, tmppanel0, &lddwork, ipiv+s*nb, &iinfo);
			if ( (*info == 0) && (iinfo > 0) )
				*info = iinfo + s*nb;

			//transfer ipiv matrix
			magma_zsetintarray_async(mindim, ipiv, 0, d_ipiv, 0, queue, &event[1]);
			magma_event_sync(event[1]);

			magma_zlaswp(n, inA(s, 0), ldda, 1, nb0, d_ipiv + s * nb, 1, queue);
			magma_kernel_sync();
			for (jj = 0; jj < nb0; jj++) {
				*(ipiv + s*nb + jj) += s*nb;
			}

			// upload i-th panel
			magma_zsetmatrix(rows, nb0, tmppanel0, 0, lddwork, inA(s, s), 0, ldda, queue);
			if((n - s * nb - nb0) > 0)
				magma_ztrsm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit, 
						nb0, n-s*nb-nb0, 
						c_one, inA(s,s),     0, ldda, 
						inA(s,s) + nb0 * ldda, 0, ldda, queue);
		}

		magma_free(d_ipiv);
		magma_free_pinned( work );
	}

	magma_queue_sync(queue);

	return *info;
	/* End of MAGMA_ZGETRF_MIC */
}

#undef inA
