/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       June 2012

       @precisions normal z -> s d c

*/
#include "common_magma.h"


extern "C" magma_int_t
magma_zpotrf_mmic(int num_mics, magma_uplo_t uplo, magma_int_t n, 
                 magmaDoubleComplex **d_lA, magma_int_t ldda, magma_int_t *info, magma_queue_t queue )
{
/*  -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       June 2012

    Purpose   
    =======   
    ZPOTRF computes the Cholesky factorization of a complex Hermitian   
    positive definite matrix dA.   

    The factorization has the form   
       dA = U**H * U,  if UPLO = 'U', or   
       dA = L  * L**H,  if UPLO = 'L',   
    where U is an upper triangular matrix and L is lower triangular.   

    This is the block version of the algorithm, calling Level 3 BLAS.   

    Arguments   
    =========   
    UPLO    (input) CHARACTER*1   
            = 'U':  Upper triangle of dA is stored;   
            = 'L':  Lower triangle of dA is stored.   

    N       (input) INTEGER   
            The order of the matrix dA.  N >= 0.   

    dA      (input/output) COMPLEX_16 array on the mic, dimension (LDDA,N)   
            On entry, the Hermitian matrix dA.  If UPLO = 'U', the leading   
            N-by-N upper triangular part of dA contains the upper   
            triangular part of the matrix dA, and the strictly lower   
            triangular part of dA is not referenced.  If UPLO = 'L', the   
            leading N-by-N lower triangular part of dA contains the lower   
            triangular part of the matrix dA, and the strictly upper   
            triangular part of dA is not referenced.   

            On exit, if INFO = 0, the factor U or L from the Cholesky   
            factorization dA = U**H * U or dA = L * L**H.   

    LDDA     (input) INTEGER   
            The leading dimension of the array dA.  LDDA >= max(1,N).
            To benefit from coalescent memory accesses LDDA must be
            dividable by 16.

    INFO    (output) INTEGER   
            = 0:  successful exit   
            < 0:  if INFO = -i, the i-th argument had an illegal value   
            > 0:  if INFO = i, the leading minor of order i is not   
                  positive definite, and the factorization could not be   
                  completed.   
    =====================================================================   */

	#define dlA(id, i, j)  (d_lA[id] + (j) * ldda + (i))
	#define dlP(id, i, j)  (d_lP[id] + (j) * ldda + (i))

	#define panel(j)  (panel + (j))
	#define tmppanel(j)  (tmppanel + (j))
	#define tmpprevpanel(j)  (tmpprevpanel + (j))

	magma_int_t     j, nb, d, id, j_local, j_local2, blkid, crosspoint, prevj, prevtrsmrows;
	magmaDoubleComplex c_one     = MAGMA_Z_ONE;
        magmaDoubleComplex  z_one = MAGMA_Z_MAKE(  1.0, 0.0 );
        magmaDoubleComplex  mz_one = MAGMA_Z_MAKE( -1.0, 0.0 );
	magmaDoubleComplex c_neg_one = MAGMA_Z_NEG_ONE;
	magmaDoubleComplex *panel, *tmppanel0, *tmppanel1, *tmppanel, *tmpprevpanel;
	double          d_one     =  1.0;
	double          d_neg_one = -1.0;
	magmaDoubleComplex *d_lP[MagmaMaxDevices], *dlpanel;
	magma_int_t rows, trsmrows, myrows, n_id, nmic, n_local[MagmaMaxDevices], lddat_local[MagmaMaxDevices], lddat, ldpanel;
	magma_event_t event[MagmaMaxDevices][4];

        double    one =  1.0;
        double  m_one = -1.0;

	*info = 0;
	if ( uplo != MagmaUpper && uplo != MagmaLower ) {
		*info = -1;
	} else if (n < 0) {
		*info = -2;
	} else if (ldda < max(1,n)) {
		*info = -4;
	}
	if (*info != 0) {
		magma_xerbla( __func__, -(*info) );
		return *info;
	}

	nb = magma_get_zpotrf_nb(n);

	ldpanel = ldda;
	if (MAGMA_SUCCESS != magma_zmalloc_pinned( &panel, 2 * nb * ldpanel )) {
		*info = MAGMA_ERR_HOST_ALLOC;
		return *info;
	}

	tmppanel0 = panel;
	tmppanel1 = tmppanel0 + nb * ldpanel;

	if ((nb <= 1) || (nb >= n)) {
		//Use unblocked code.
		magma_setdevice(0);
		magma_zgetmatrix( n, n, dlA(0, 0, 0), 0, ldda, panel, 0, ldpanel, queue);
		lapackf77_zpotrf( lapack_const(uplo), &n, panel, &ldpanel, info);
		magma_zsetmatrix( n, n, panel, 0, ldpanel, dlA(0, 0, 0), 0, ldda, queue );
	} else {

		for( d = 0; d < num_mics; d++ ) {
			magma_setdevice(d);

			// local-n and local-ld 
			n_local[d] = ((n / nb) / num_mics) * nb;
			if (d < (n / nb) % num_mics)
				n_local[d] += nb;
			else if (d == (n / nb) % num_mics)
				n_local[d] += n % nb;

			lddat_local[d] = ((n_local[d] + 31) / 32) * 32;
			lddat_local[d] = (lddat_local[d] % 256 == 0) ? lddat_local[d] + 32 : lddat_local[d];

			if (MAGMA_SUCCESS != magma_zmalloc( &d_lP[d], nb * ldda )) {
				for( j = 0; j < d; j++ ) {
					magma_setdevice(j);
					magma_free( d_lP[d] );
				}
				*info = MAGMA_ERR_DEVICE_ALLOC;
				return *info;
			}

			for( j = 0; j < 4; j++ ) {
				event[d][j] = NULL;
			}
		}

		//#define ENABLE_TIMER
		#if defined (ENABLE_TIMER)
		double therk[4], tmtc, tcchol, tctrsm, tctm, tmnp, tcnp;
		printf("\n\n %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s\n", 
				"j", "nb", "row", "mtc", "CPU_np", "panel", "ctrsm", "CH+TRSM", "CPU", "dsyrk[0]", "dsyrk[1]", "dsyrk[2]", "dsyrk[3]", "ctm P", "MIC_np");
		printf("     ====================================================================================================\n");
                #endif 

		//Use blocked code.
		if (uplo == MagmaUpper) 
		{    
		} else {

			blkid = -1;
			// Compute the Cholesky factorization A = L*L' 
			for (j = 0; (j + nb) < n; j += nb) {

                        	#if defined (ENABLE_TIMER)
                        	therk[0] = therk[1] = therk[2] = therk[3] = tmtc = tcchol = tctrsm = tctm = tmnp = tcnp = 0.0;
                        	#endif

				blkid += 1;
				tmppanel = (blkid % 2 == 0) ? tmppanel0 : tmppanel1;
				// Set the mic number that holds the current panel 
				id = (j / nb) % num_mics;
				magma_setdevice(id);

				//Set the local index where the current panel is
				j_local = j / (nb * num_mics) * nb;
				
				rows = n - j;
				//factor the diagonal
				magma_setdevice(id);
				
				magma_kernel_sync();

                                #if defined (ENABLE_TIMER)
				tmtc = get_time();
                                #endif

				// send the diagonal to cpu 
				magma_zgetmatrix_async( rows, nb,
						dlA(id, j, j_local), 0, ldda,
						tmppanel(j), 0, ldpanel, queue, &event[id][0] );
				
				magma_event_sync( event[id][0] );

                                #if defined (ENABLE_TIMER)
				tmtc = get_time() - tmtc;
                                #endif

                                #if defined (ENABLE_TIMER)
                                tcchol = get_time();
                                #endif

				lapackf77_zpotrf(MagmaLowerStr, &nb, tmppanel(j), &ldpanel, info);
				if (*info != 0) {
					*info = *info + j;
					//break;
				}

                                #if defined (ENABLE_TIMER)
                                tcchol = get_time() - tcchol;
                                #endif

                                #if defined (ENABLE_TIMER)
                                tctm = get_time();
                                #endif

				//send the diagonal to mics
				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_kernel_sync();

					if( d == id ) {
						dlpanel = dlA(d, j, j_local);
					} else {
						dlpanel = dlP(d, 0, 0);
					}

					magma_zsetmatrix_async(rows, nb,
							tmppanel(j),    0, ldpanel,
							dlpanel, 0, ldda, queue, &event[d][1] );
				}

                                #if defined (ENABLE_TIMER)
				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_event_sync( event[d][1] );
				}

				tctm = get_time() - tctm;
                                #endif

				#if defined (ENABLE_TIMER)
                                tctrsm = get_time();
                                #endif

				trsmrows = rows - nb;

				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_event_sync( event[d][1] );

					if( d == id ) {
						dlpanel = dlA(d, j, j_local);
					} else {
						dlpanel = dlP(d, 0, 0);
					}

					if(trsmrows > 0) {
						magma_ztrsm(MagmaRight, MagmaLower, MagmaConjTrans, MagmaNonUnit, trsmrows, nb, z_one, dlpanel, 0, ldda, dlpanel + nb, 0, ldda, queue);
					}
				}

                                #if defined (ENABLE_TIMER)
				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_kernel_sync();
				}
                                tctrsm = get_time() - tctrsm;
                                #endif


				if ( (j + nb) < n) {
					magma_int_t offset = j + nb;

					//syrk on multiple mic
					d = (id + 1) % num_mics;
					for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
						magma_setdevice(d);
						magma_kernel_sync();

						//update the entire trailing matrix
						if( d == id ) {
							dlpanel = dlA(d, j + nb, j_local);
						} else {
							dlpanel = dlP(d, nb, 0);
						}

                                		#if defined (ENABLE_TIMER)
						therk[d] = get_time();
                                		#endif

						if(n - offset > 0) {
							magma_zherk_mmic(MagmaLower, MagmaTrans, n - offset, nb,
									m_one, dlpanel, ldda, 0,
									one, d_lA[d], ldda,  offset,
									num_mics, nb, d);
						}

                                		#if defined (ENABLE_TIMER)
						magma_kernel_sync();
						therk[d] = get_time() - therk[d];
                                		#endif
					}

					#if defined (ENABLE_TIMER)
					printf("%10d %10d %10d %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf(%d)\n",
							j, nb, rows, tmtc * 1000, tcnp*1000, tcchol*1000, tctrsm * 1000, (tcchol + tctrsm) * 1000, 
							(tmtc+tcnp+tcchol+tctrsm)*1000, therk[0] * 1000, therk[1] * 1000, therk[2] * 1000, therk[3] * 1000, tctm * 1000,  
							tmnp * 1000, (id + 1) % num_mics);
					fflush(0);
					#endif
	

				}
			}

			//cholesky for the last block
			if(j < n) {
				rows = n - j;
				id = (j / nb) % num_mics;
				magma_setdevice(id);
				magma_kernel_sync();

				//Set the local index where the current panel is
				j_local = j / (nb * num_mics) * nb;
				
				magma_zgetmatrix(rows, rows, dlA(id, j, j_local), 0, ldda, panel(j), 0, ldpanel, queue);
				lapackf77_zpotrf(MagmaLowerStr, &rows, panel(j), &ldpanel, info);
				magma_zsetmatrix(rows, rows, panel(j), 0, ldpanel, dlA(id, j, j_local), 0, ldpanel, queue);
			}

		} // end of else not upper

		//clean up
		for( d = 0; d < num_mics; d++ ) {
			magma_setdevice(d);
			magma_free( d_lP[d] );
		}

	} // end of not lapack

	// free workspace
	magma_free_pinned( panel );

	return *info;
} /* magma_zpotrf_mmic */
