/*
   -- micMAGMA (version 1.4.0) --
   Univ. of Tennessee, Knoxville
   Univ. of California, Berkeley
   Univ. of Colorado, Denver
   @date July 2015

 @author Khairul Kabir
 @author Azzam Haidar

 @generated from zpotrf_mic.cpp normal z -> c, Sun Jul 12 17:46:26 2015

*/

#include "common_magma.h"

extern "C" magma_int_t
magma_cpotrf_mic( magma_uplo_t   uplo, magma_int_t    n,
                  magmaFloatComplex_ptr dA, size_t dA_offset, magma_int_t ldda,
                  magma_int_t*   info, magma_queue_t queue )
{
/*  -- micMAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

    Purpose
    =======
    CPOTRF computes the Cholesky factorization of a complex Hermitian
    positive definite matrix dA.

    The factorization has the form
       dA = U\*\*H * U,  if UPLO = 'U', or
       dA = L  * L\*\*H,  if UPLO = 'L',
    where U is an upper triangular matrix and L is lower triangular.

    This is the block version of the algorithm, calling Level 3 BLAS.

    Arguments
    =========
    UPLO    (input) INTEGER
            = MagmaUpper:  Upper triangle of dA is stored;
            = MagmaLower:  Lower triangle of dA is stored.

    N       (input) INTEGER
            The order of the matrix dA.  N >= 0.

    dA      (input/output) COMPLEX array on the GPU, dimension (LDDA,N)
            On entry, the Hermitian matrix dA.  If UPLO = 'U', the leading
            N-by-N upper triangular part of dA contains the upper
            triangular part of the matrix dA, and the strictly lower
            triangular part of dA is not referenced.  If UPLO = 'L', the
            leading N-by-N lower triangular part of dA contains the lower
            triangular part of the matrix dA, and the strictly upper
            triangular part of dA is not referenced.

            On exit, if INFO = 0, the factor U or L from the Cholesky
            factorization dA = U\*\*H*U or dA = L*L\*\*H.

    LDDA    (input) INTEGER
            The leading dimension of the array dA.  LDDA >= max(1,N).
            To benefit from coalescent memory accesses LDDA must be
            dividable by 16.

    INFO    (output) INTEGER
            = 0:  successful exit
            < 0:  if INFO = -i, the i-th argument had an illegal value
            > 0:  if INFO = i, the leading minor of order i is not
                  positive definite, and the factorization could not be
                  completed.
    =====================================================================   */

    #define dA(a_1, a_2)    (dA + (a_1) + (a_2)*(ldda))
    #define tmppanel(i_)    (tmppanel + (i_))
    #define tmpprevpanel(i_)    (tmpprevpanel + (i_))

    magmaFloatComplex c_one  = MAGMA_C_ONE;
    magmaFloatComplex c_zero = MAGMA_C_ZERO;
    magmaFloatComplex  z_one = MAGMA_C_MAKE(  1.0, 0.0 );
    magmaFloatComplex mz_one = MAGMA_C_MAKE( -1.0, 0.0 );

    float    one =  1.0;
    float  m_one = -1.0;

    magmaFloatComplex  *work, *hwork;
    magmaFloatComplex  *tmppanel, *tmppanel0, *tmppanel1, *tmpprevpanel;

    magma_int_t i, ldwork, lddwork, old_i, rows, old_rows, trsmrows, old_trsmrows, blkid;
    magma_int_t crosspoint;
    magma_int_t nbmin, nx, nb;
    magma_int_t lwork, lhtmppanel;
    magma_int_t numblk, min_nb, nb_max, curr_nb, prev_nb, next_nb;
    magma_event_t event[2] = {NULL, NULL};
    magma_int_t nb0 = 480, nb1 = 720, nb2 = 960, nb3 = 1440, nb4 = 1680, nb5 = 1920;

    *info = 0;
    if ( uplo != MagmaUpper && uplo != MagmaLower ) {
        *info = -1;
    } else if ( n < 0 ) {
        *info = -2;
    } else if ( ldda < max(1,n) ) {
        *info = -4;
    }
    if ( *info != 0 ) {
        magma_xerbla( __func__, -(*info) );
        return *info;
    }

    if (n >= 31680)
        nb_max = nb5;
    else if (n >= 24960)
        nb_max = nb4;
    else if (n >= 19680)
        nb_max = nb3;
    else if (n >= 13440)
        nb_max = nb2;
    else if (n >= 9120)
        nb_max = nb1;
    else
        nb_max = nb0;

    nb = min_nb = 480;
    numblk = n / min_nb;

    // CPU  will  store panel in two consecutive temporary panel(size=2*n*nb_max)
    lhtmppanel  = 2 * n * nb_max;

    if ( MAGMA_SUCCESS != magma_cmalloc_pinned( &work, lhtmppanel ) ) {
        *info = MAGMA_ERR_HOST_ALLOC;
        return *info;
    }

    ldwork = n;
    // setup pointer
    tmppanel0 = work;
    tmppanel1 = work + n * nb_max;

    if ( n > min_nb ) {
        // crosspoint = 0 mean that next panel update is done on CPU all the time
        // crosspoint = N mean that next panel update is done on MIC all the time
        crosspoint  = 0; //4320;

        //#define ENABLE_TIMER
        #if defined (ENABLE_TIMER)
        float therk, tmtc, tcchol, tctrsm, tctmp, tmnp, tcnp;
        printf("\n\n %10s %10s %10s %10s  %10s %10s %10s %10s  %10s  %10s %10s  %10s\n",
                "i", "curr_nb", "row", "panel", "ctrsm", "CH+TRSM", "CPU", "dsyrk", "CPU_np", "MIC_np", "ctm P", "mtc");
        printf("     ====================================================================================================\n");
        #endif

        /* Use blocked code initially */

        old_i = 0; nb = min(n, nb); old_rows = n;
        curr_nb = next_nb = min_nb;
        blkid = -1;

        for (i = 0; (i + next_nb) < n; i += curr_nb) {

            magma_kernel_sync();
            blkid += 1;
            rows  = n - i;
            tmppanel = blkid%2 == 0 ? tmppanel0 : tmppanel1;


            #if defined (ENABLE_TIMER)
            therk = tmtc = tcchol = tctrsm = tctmp = tmnp = tcnp = 0.0;
            #endif

            curr_nb = next_nb;

            /*
            if ((rows - curr_nb) >= 37440)
                next_nb = 1920;
            else if ((rows - curr_nb) >= 30720)
                next_nb = 1680;
            else if ((rows - curr_nb) >= 22640)
                next_nb = 1440;
            else if ((rows - curr_nb) >= 13440)
                next_nb = 960;
            else if ((rows - curr_nb) >= 9120)
                next_nb = 720;
            else
                next_nb = 480;
            */

            if ((rows - curr_nb) >= 31680) {
                if (blkid == 0) next_nb = nb5 - 480;
                else next_nb = nb5;
            } else if ((rows - curr_nb) >= 27840) {
                if (blkid == 0) next_nb = nb4 - 480;
                else next_nb = nb4;
            } else if ((rows - curr_nb) >= 20160) {
                if (blkid == 0) next_nb = nb3 - 480;
                else next_nb = nb3;
            } else if ((rows - curr_nb) >= 13440) {
                next_nb = nb2;
            } else if ((rows - curr_nb) >= 9120) {
                next_nb = nb1;
            } else {
                next_nb = nb0;
            }
    
            if (i > 0) {
                #if defined (ENABLE_TIMER)
                therk = magma_wtime();
                #endif
                
                magma_cherk( MagmaLower, MagmaNoTrans,
                        old_trsmrows - curr_nb, prev_nb,
                        m_one, dA(old_i + prev_nb + curr_nb, old_i ), 0, ldda,
                        one, dA(old_i + prev_nb + curr_nb, old_i + prev_nb + curr_nb), 0, ldda, queue);

                #if defined (ENABLE_TIMER)
                magma_kernel_sync();
                therk = magma_wtime() - therk;
                #endif

                tmpprevpanel = (blkid - 1)%2 == 0 ? tmppanel0 : tmppanel1;
                // CPU do the next panel, bring it from the MIC
                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime();
                #endif

                magma_cgetmatrix_async(rows, curr_nb, dA(i, i), 0, ldda, tmppanel(i), 0, ldwork, queue, &event[0]);
                magma_event_sync(event[0]);

                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime() - tmtc;
                #endif
                //CPU do the look ahead panel first then factorize it
                if (old_rows > crosspoint) {
                    #if defined (ENABLE_TIMER)
                    tcnp = magma_wtime();
                    #endif
                    blasf77_cgemm( MagmaNoTransStr, MagmaConjTransStr,
                            &rows, &curr_nb, &prev_nb,
                            &mz_one, tmpprevpanel(i), &ldwork, tmpprevpanel(i), &ldwork,
                            &z_one, tmppanel(i), &ldwork);

                    #if defined (ENABLE_TIMER)
                    tcnp = magma_wtime() - tcnp;
                    #endif
                }

            } else {
                // i=0 ==> first time
                // or MIC was updating the next panel and so need to copy it to CPU
                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime();
                #endif
                   
                magma_cgetmatrix_async(rows, curr_nb, dA(i, i), 0, ldda, tmppanel(i), 0, ldwork, queue, &event[0]);
                magma_event_sync(event[0]);
                   
                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime() - tmtc;
                #endif
            }


            if (rows <= 3360)
                crosspoint = 0;
            else if (n <= 17240 && blkid == 0)
                crosspoint = n;
            else
                crosspoint = 5280;

            #if defined (ENABLE_TIMER)
            tcchol = magma_wtime();
            #endif

            lapackf77_cpotrf(MagmaLowerStr, &curr_nb, tmppanel(i), &ldwork, info);

            if ( *info != 0 ) {
                assert( *info > 0 );
                *info += i;
                break;
            }

            #if defined (ENABLE_TIMER)
            tcchol = magma_wtime() - tcchol;
            #endif

            #if defined (ENABLE_TIMER)
            tctrsm = magma_wtime();
            #endif

            trsmrows=rows - curr_nb;
            if (trsmrows > 0) {
                blasf77_ctrsm(MagmaRightStr, MagmaLowerStr, MagmaConjTransStr, MagmaNonUnitStr, &trsmrows, &curr_nb, &z_one, tmppanel(i), &ldwork, tmppanel(i+curr_nb), &ldwork);
            }

            #if defined (ENABLE_TIMER)
            tctrsm = magma_wtime() - tctrsm;
            #endif

            #if defined (ENABLE_TIMER)
            tctmp = magma_wtime();
            #endif

            // ======================================================
            //            SEND panel to MIC
            // ======================================================
            magma_csetmatrix_async(rows, curr_nb, tmppanel(i), 0, ldwork, dA(i,i), 0, ldda, queue, &event[0]);
            magma_event_sync(event[0]);

            #if defined (ENABLE_TIMER)
            tctmp = magma_wtime() - tctmp;
            #endif
            
            // CPU UPDATE NEXT PANEL if rows > crosspoint
            // rows > crosspoint:   CPU do next panel then send R
            // rows <= crosspoint: MIC send R then do next panel

            if ((i + curr_nb + next_nb) < n) {
                if (rows <= crosspoint) {
                    #if defined (ENABLE_TIMER)
                    tmnp = magma_wtime();
                    #endif
                    // MIC is WORKING on next panel
                    magma_cgemm( MagmaNoTrans, MagmaConjTrans,
                            trsmrows, next_nb, curr_nb,
                            mz_one, dA(i + curr_nb, i), 0, ldda, dA(i + curr_nb, i), 0, ldda,
                            z_one, dA(i + curr_nb, i + curr_nb), 0, ldda, queue);

                    magma_kernel_sync();

                    #if defined (ENABLE_TIMER)
                    tmnp = magma_wtime() - tmnp;
                    #endif
                }
            }
            else {
                magma_cherk( MagmaLower, MagmaNoTrans,
                        trsmrows, curr_nb,
                        m_one, dA(i + curr_nb, i   ), 0, ldda,
                        one,   dA(i + curr_nb, i + curr_nb), 0, ldda, queue);
                magma_kernel_sync();
            }

            old_i  = i;
            old_rows = rows;
            old_trsmrows = trsmrows;

            prev_nb = curr_nb;

            #if defined (ENABLE_TIMER)
            printf("%10d %10d %10d %10.3lf  %10.3lf  %10.3lf %10.3lf %10.3lf  %10.3lf %10.3lf  %10.3lf %10.3lf\n",
                    i, curr_nb, rows, tcchol * 1000, tctrsm*1000, (tcchol+tctrsm)*1000, (tcchol+tctrsm+tcnp+tmtc+tctmp)*1000,
                    therk * 1000, tcnp * 1000, tmnp * 1000, tctmp * 1000, tmtc * 1000);
            #endif

        }
    } else {
        i = 0;
    }

    /* Use unblocked code to factor the last or only block. */
    if (i < n) {
        rows = n - i;
        magma_cgetmatrix(rows, rows, dA(i, i), 0, ldda, tmppanel0, 0, ldwork, queue);
        lapackf77_cpotrf(MagmaLowerStr, &rows, tmppanel0, &ldwork, info);
        magma_csetmatrix(rows, rows, tmppanel0, 0, ldwork, dA(i, i), 0, ldda, queue);
    }

    magma_free_pinned(work);

    return *info;
} /* magma_cgeqrf2_mic */
