/*
   -- micMAGMA (version 1.4.0) --
   Univ. of Tennessee, Knoxville
   Univ. of California, Berkeley
   Univ. of Colorado, Denver
   @date July 2015

   @generated from zgetrf_mic.cpp normal z -> d, Sun Jul 12 17:46:26 2015
*/

#include "common_magma.h"

extern "C" magma_int_t
magma_dgetrf_mic(magma_int_t m, magma_int_t n,
                 magmaDouble_ptr dA, size_t dA_offset, magma_int_t ldda,
                 magma_int_t *ipiv, magma_int_t *info,
                 magma_queue_t queue )
{
/*  -- micMAGMA (version 1.4.0) --
    Univ. of Tennessee, Knoxville
    Univ. of California, Berkeley
    Univ. of Colorado, Denver
    @date July 2015

    Purpose
    =======
    DGETRF computes an LU factorization of a general M-by-N matrix A
    using partial pivoting with row interchanges.
    
    The factorization has the form
    A = P * L * U
    where P is a permutation matrix, L is lower triangular with unit
    diagonal elements (lower trapezoidal if m > n), and U is upper
    triangular (upper trapezoidal if m < n).

    This is the right-looking Level 3 BLAS version of the algorithm.
    
    Arguments
    =========
    
    M       (input) INTEGER
            The number of rows of the matrix A.  M >= 0.
    
    N       (input) INTEGER
            The number of columns of the matrix A.  N >= 0.

    A       (input/output) DOUBLE_PRECISION array on the GPU, dimension (LDDA,N).
            On entry, the M-by-N matrix to be factored.
            On exit, the factors L and U from the factorization
            A = P*L*U; the unit diagonal elements of L are not stored.

    LDDA   (input) INTEGER
           The leading dimension of the array A.  LDDA >= max(1,M).

    IPIV   (output) INTEGER array, dimension (min(M,N))
           The pivot indices; for 1 <= i <= min(M,N), row i of the
           matrix was interchanged with row IPIV(i).
       
    INFO   (output) INTEGER
           = 0:  successful exit
           < 0:  if INFO = -i, the i-th argument had an illegal value
                 if INFO = -7, internal GPU memory allocation failed.
           > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
                 has been completed, but the factor U is exactly
             singular, and division by zero will occur if it is used
             to solve a system of equations.
=====================================================================    */

    #define inA(i,j) (dA + dA_offset + (j) * ldda + (i))
    #define tmppanel(i_) (tmppanel + (i_))
    #define tmpprevpanel(i_) (tmpprevpanel + (i_))

    double c_one     = MAGMA_D_MAKE(  1.0, 0.0 );
    double c_neg_one = MAGMA_D_MAKE( -1.0, 0.0 );

    magma_int_t iinfo, nb;
    magma_int_t mindim;
    magma_int_t i, rows, s, cols, lddwork;

    magma_int_t *d_ipiv;

    double *work, *tmppanel0, *tmppanel1, *tmppanel, *tmpprevpanel;
    magma_int_t nb_max, nb_min, curr_nb, prev_nb, blkid, next_nb, old_rows, crosspoint;

    magma_int_t err, nb0 = 480, nb1 = 960, nb2 = 1200, nb3 = 1680;

    *info = 0;
    int jj;

    if (m < 0)
        *info = -1;
    else if (n < 0)
        *info = -2;
    else if (ldda < max(1,m))
        *info = -4;

    if (*info != 0) {
        magma_xerbla( __func__, -(*info) );
        return *info;
    }

    if (m == 0 || n == 0)
        return MAGMA_SUCCESS;

    mindim = min(m, n);
    
    nb = nb_min = nb0;
    
    if (m >= 27360 )
        nb_max = nb3;
    else if (m >= 18720)
        nb_max = nb2;
    else if (m >= 9120)
        nb_max = nb1;
    else
        nb_max = nb0;

    //nb     = magma_get_dgetrf_nb(m);
    
    s      = mindim / nb_min;

    magma_event_t event[2] = {NULL, NULL};
    crosspoint = 0; //5280;

    if (nb < 1 || nb >= min(m,n)) {
        // use CPU code
        err = magma_dmalloc_pinned( &work, m*n );
        if ( err != MAGMA_SUCCESS ) {
            *info = MAGMA_ERR_HOST_ALLOC;
            return *info;
        }

        chk( magma_dgetmatrix( m, n, dA, dA_offset, ldda, work, 0, m, queue ));
        lapackf77_dgetrf(&m, &n, work, &m, ipiv, info);
        chk( magma_dsetmatrix( m, n, work, 0, m, dA, dA_offset, ldda, queue ));

        magma_free_pinned(work);
    }
    else {
        // use hybrid blocked code
        lddwork = ldda;

        //memory allocate for d_ipiv
        if ( MAGMA_SUCCESS != magma_imalloc( &d_ipiv, ldda )) {
            *info = MAGMA_ERR_DEVICE_ALLOC;
            return *info;
        }

        if ( MAGMA_SUCCESS != magma_dmalloc_pinned( &work, (2 * ldda * max(nb_max, n - s * nb_min)) ) ) {
            *info = MAGMA_ERR_HOST_ALLOC;
            return *info;
        }

        //#define ENABLE_TIMER
        #if defined (ENABLE_TIMER)
        double ttrsmgemm, ttrsm, tgemm, tmtc, tipiv, tgetrf, tswap, tswap0, tctm, tcnp, tmnp, tnp1, tnp2;
        printf("\n\n %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s\n",
            "i", "curr_nb", "row", "panel", "tipiv", "CPU_np", "np mtcp", "np ctm", "CPU", "trsmgemm", "mtswap0", "mtswap1", "MIC_np", "ctm P", "mtc");
        printf("     ====================================================================================================\n");
        #endif

        tmppanel0 = work;
        tmppanel1 = work + nb_max * ldda;
        
        curr_nb = prev_nb = nb_min;

        old_rows = 0;
        blkid = -1;
        next_nb = nb_min;

        for (i = 0; (i + next_nb) < (s * nb_min); i += curr_nb)
        {
            rows = m - i;
            blkid += 1;

            tmppanel = (blkid % 2) == 0 ? tmppanel0 : tmppanel1;

            #if defined (ENABLE_TIMER)
            ttrsmgemm = tmtc = tgetrf = tswap = tswap0 = tctm = tmnp = tcnp = tipiv = 0.0;
            #endif

            curr_nb = next_nb;

            magma_int_t changeNBU, changeNBD;

            if ((rows - curr_nb) >= 27360) {
                if (blkid == 0) next_nb = nb3 - 480;
                else next_nb = nb3;
            } else if ((rows - curr_nb) >= 18720 ) {
                if (blkid == 0) next_nb = nb2 - 240;
                else next_nb = nb2;
            } else if ((rows - curr_nb) >= 9120 ) {
                next_nb = nb1;
            } else {
                next_nb = nb0;
            }
            
            if ( i > 0 ) {

                tmpprevpanel = ((blkid - 1) % 2) == 0 ? tmppanel0 : tmppanel1;
                
                #if defined (ENABLE_TIMER)
                ttrsmgemm = magma_wtime();
                #endif

                magma_dtrsmgemm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit,
                        prev_nb, n - i - curr_nb,
                        c_one, inA(i - prev_nb, i - prev_nb), 0, ldda,
                        inA(i - prev_nb, i + curr_nb), 0, ldda,
                        MagmaNoTrans, MagmaNoTrans,
                        m - i, n - i - curr_nb, prev_nb,
                        c_neg_one, inA(i, i - prev_nb), 0, ldda,
                        inA(i - prev_nb,  i + curr_nb), 0, ldda,
                        c_one,     inA(i,  i + curr_nb), 0, ldda, queue );

                    
                #if defined (ENABLE_TIMER)
                magma_kernel_sync();
                ttrsmgemm = magma_wtime() - ttrsmgemm;
                #endif

                if (old_rows > crosspoint ) {

                    #if defined (ENABLE_TIMER)
                    tnp1 = magma_wtime();
                    #endif

                    //donwload look ahead panel
                    magma_dgetmatrix_async(m - i + prev_nb, curr_nb, inA(i - prev_nb, i), 0, ldda, tmppanel(i - prev_nb), 0, lddwork, queue, &event[0]);
                    magma_event_sync(event[0]);

                    #if defined (ENABLE_TIMER)
                    tnp1 = magma_wtime() - tnp1;
                    #endif

                    #if defined (ENABLE_TIMER)
                    tcnp = magma_wtime();
                    #endif

                    rows = m - i;
                    dtrsm_( "L", "L", "N", "U",
                            &prev_nb, &curr_nb,
                            &c_one, tmpprevpanel(i - prev_nb), &lddwork,
                            tmppanel(i - prev_nb), &lddwork);

                    dgemm_( "N", "N",
                            &rows, &curr_nb, &prev_nb,
                            &c_neg_one, tmpprevpanel(i), &lddwork,
                            tmppanel(i - prev_nb), &lddwork,
                            &c_one,     tmppanel(i), &lddwork );

                    #if defined (ENABLE_TIMER)
                    tcnp = magma_wtime() - tcnp;
                    #endif

                    #if defined (ENABLE_TIMER)
                    tnp2 = magma_wtime();
                    #endif
                    //send prev_nb X curr_nb trsm part back to mic
                    magma_dsetmatrix_async(prev_nb, curr_nb, tmppanel(i - prev_nb), 0, lddwork, inA(i - prev_nb, i), 0, ldda, queue, &event[1]);
                    magma_event_sync(event[1]);

                    #if defined (ENABLE_TIMER)
                    tnp2 = magma_wtime() - tnp2;
                    #endif
                }
            }
            
            if (i == 0 || old_rows <= crosspoint) {
                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime();
                #endif

                // download i-th panel
                magma_dgetmatrix_async(m - i, curr_nb, inA(i, i), 0, ldda, tmppanel(i), 0, lddwork, queue, &event[0]);
                magma_event_sync(event[0]);

                #if defined (ENABLE_TIMER)
                tmtc = magma_wtime() - tmtc;
                #endif
            }

            #if defined (ENABLE_TIMER)
            tgetrf = magma_wtime();
            #endif

            // do the cpu part
            rows = m - i;
            lapackf77_dgetrf( &rows, &curr_nb, tmppanel(i), &lddwork, ipiv + i, &iinfo);
            //if ( (*info == 0) && (iinfo > 0) )
            //    *info = iinfo + i;

            #if defined (ENABLE_TIMER)
            tgetrf = magma_wtime() - tgetrf;
            #endif
                
            #if defined (ENABLE_TIMER)
            tipiv = magma_wtime();
            #endif

            //transfer ipiv
            magma_dsetintarray_async(mindim, ipiv, 0, d_ipiv, 0, queue, &event[1]);
            magma_event_sync(event[1]);

            for (jj = 0; jj < curr_nb; jj++) {
                *(ipiv + i + jj) += i;
            }

            #if defined (ENABLE_TIMER)
            tipiv = magma_wtime() - tipiv;
            #endif

            #if defined (ENABLE_TIMER)
            tctm = magma_wtime();
            #endif

            // upload i-th panel
            magma_dsetmatrix_async(m - i, curr_nb, tmppanel(i), 0, lddwork, inA(i, i), 0, ldda, queue, &event[1]);

            #if defined (ENABLE_TIMER)
            magma_event_sync(event[1]);
            tctm = magma_wtime() - tctm;
            #endif
            
            #if defined (ENABLE_TIMER)
            tswap0 = magma_wtime();
            #endif
        
            magma_dlaswp(i, inA(i, 0), ldda, 1, curr_nb, d_ipiv + i, 1, queue);

            #if defined (ENABLE_TIMER)
            magma_kernel_sync();
            tswap0 = magma_wtime() - tswap0;
            #endif

            #if defined (ENABLE_TIMER)
            tswap = magma_wtime();
            #endif
            
            magma_dlaswp(n - i - curr_nb, inA(i, i + curr_nb), ldda, 1, curr_nb, d_ipiv + i, 1, queue);


            #if defined (ENABLE_TIMER)
            magma_kernel_sync();
            tswap = magma_wtime() - tswap;
            #endif

            magma_event_sync(event[1]);
            magma_kernel_sync();

            if (blkid == 0)
                crosspoint = n;
            else
                crosspoint = 0;

            //mic will do just first look ahead update update
            if ((i + curr_nb + next_nb) < (s * nb_min)) {
                if (rows <= crosspoint) {
                    #if defined (ENABLE_TIMER)
                    tmnp = magma_wtime();
                    #endif

                    magma_dtrsmgemm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit,
                            curr_nb, next_nb,
                            c_one, inA(i, i), 0, ldda,
                            inA(i, i + curr_nb), 0, ldda,
                            MagmaNoTrans, MagmaNoTrans,
                            m - i - curr_nb, next_nb, curr_nb,
                            c_neg_one, inA(i + curr_nb, i), 0, ldda,
                            inA(i,  i + curr_nb), 0, ldda,
                            c_one,     inA(i + curr_nb,  i + curr_nb), 0, ldda, queue );

                    magma_kernel_sync();

                    #if defined (ENABLE_TIMER)
                    tmnp = magma_wtime() - tmnp;
                    #endif
                }
            } else {
                // non-parallel computations -- last block
                magma_dtrsmgemm( MagmaLeft, MagmaLower, MagmaNoTrans, MagmaUnit,
                        curr_nb, n - (i + curr_nb),
                        c_one, inA(i, i), 0, ldda,
                        inA(i, i+curr_nb), 0, ldda,
                        MagmaNoTrans, MagmaNoTrans,
                        m - (i + curr_nb), n - (i + curr_nb), curr_nb,
                        c_neg_one, inA(i + curr_nb,   i), 0, ldda,
                        inA(i, i+curr_nb  ), 0, ldda,
                        c_one,     inA(i+curr_nb, i+curr_nb), 0, ldda, queue );
        
            }

            prev_nb = curr_nb;
            old_rows = rows;

            #if defined (ENABLE_TIMER)
            printf("%10d %10d %10d %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf % 10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf\n",
                    i, curr_nb, rows, tgetrf * 1000, tipiv*1000, tcnp * 1000, tnp1 * 1000, tnp2 * 1000, (tgetrf + tipiv + tcnp + tctm + tnp1 + tnp2)*1000, ttrsmgemm * 1000,
                    tswap0 * 1000, tswap * 1000, tmnp * 1000, tctm * 1000, tmtc * 1000);
            #endif

        }

        magma_int_t nb0 = min(m - i, n - i);
        
        if (nb0 > 0) {
            rows = m - i;
            cols = n - i;
            magma_dgetmatrix(rows, cols, inA(i, i), 0, ldda, tmppanel0, 0, lddwork, queue);

            // do the cpu part
            lapackf77_dgetrf( &rows, &cols, tmppanel0, &lddwork, ipiv+i, &iinfo);
            if ( (*info == 0) && (iinfo > 0) )
                *info = iinfo + i;

            //transfer ipiv matrix
            magma_dsetintarray_async(mindim, ipiv, 0, d_ipiv, 0, queue, &event[1]);
            magma_event_sync(event[1]);

            magma_dlaswp(n, inA(i, 0), ldda, 1, nb0, d_ipiv + i, 1, queue);
            magma_kernel_sync();
            for (jj = 0; jj < nb0; jj++) {
                *(ipiv + i + jj) += i;
            }

            // upload i-th panel
            magma_dsetmatrix(rows, cols, tmppanel0, 0, lddwork, inA(i, i), 0, ldda, queue);
        }

        magma_free(d_ipiv);
        magma_free_pinned( work );
    }

    magma_queue_sync(queue);

    return *info;
    /* End of MAGMA_DGETRF_MIC */
}

#undef inA
