/*
    -- MAGMA (version 1.4.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date July 2015

       @generated from zgesv_rbt.cpp normal z -> s, Sun Jul 12 17:46:26 2015

*/
#include "common_magma.h"
#include "magma.h"





#define BWDMAX 1.0
#define ITERMAX 30

/**
    Purpose
    -------
    SGERFS  improve the computed solution to a system of linear
          equations.

        
    The iterative refinement process is stopped if
        ITER > ITERMAX
    or for all the RHS we have:
        RNRM < SQRT(N)*XNRM*ANRM*EPS*BWDMAX
    where
        o ITER is the number of the current iteration in the iterative
          refinement process
        o RNRM is the infinity-norm of the residual
        o XNRM is the infinity-norm of the solution
        o ANRM is the infinity-operator-norm of the matrix A
        o EPS is the machine epsilon returned by SLAMCH('Epsilon')
    The value ITERMAX and BWDMAX are fixed to 30 and 1.0D+00 respectively.

    Arguments
    ---------
    @param[in]
    trans   magma_trans_t
            Specifies the form of the system of equations:
      -     = MagmaNoTrans:    A    * X = B  (No transpose)
      -     = MagmaTrans:      A**T * X = B  (Transpose)
      -     = MagmaConjTrans:  A**H * X = B  (Conjugate transpose)

    @param[in]
    n       INTEGER
            The number of linear equations, i.e., the order of the
            matrix A.  N >= 0.

    @param[in]
    nrhs    INTEGER
            The number of right hand sides, i.e., the number of columns
            of the matrix B.  NRHS >= 0.

    @param[in]
    dA      REAL array on the MIC, dimension (ldda,N)
            the N-by-N coefficient matrix A.
            
    @param[in]
    ldda    INTEGER
            The leading dimension of the array dA.  ldda >= max(1,N).

    @param[in]
    dB      REAL array on the MIC, dimension (lddb,NRHS)
            The N-by-NRHS right hand side matrix B.

    @param[in]
    lddb    INTEGER
            The leading dimension of the array dB.  lddb >= max(1,N).

    @param[in, out]
    dX      REAL array on the MIC, dimension (lddx,NRHS)
            On entry, the solution matrix X, as computed by
            SGETRS_NOPIV.  On exit, the improved solution matrix X.

    @param[in]
    lddx    INTEGER
            The leading dimension of the array dX.  lddx >= max(1,N).

    @param
    dworkd  (workspace) REAL array on the MIC, dimension (N*NRHS)
            This array is used to hold the residual vectors.

    @param
    dAF     REAL array on the MIC, dimension (ldda,n)
            The factors L and U from the factorization A = L*U
            as computed by SGETRF_NOPIV.

    @param[out]
    iter    INTEGER
      -     < 0: iterative refinement has failed, real
                 factorization has been performed
        +        -1 : the routine fell back to full precision for
                      implementation- or machine-specific reasons
        +        -2 : narrowing the precision induced an overflow,
                      the routine fell back to full precision
        +        -3 : failure of SGETRF
        +        -31: stop the iterative refinement after the 30th iteration
      -     > 0: iterative refinement has been successfully used.
                 Returns the number of iterations
 
    @param[out]
    info   INTEGER
      -     = 0:  successful exit
      -     < 0:  if info = -i, the i-th argument had an illegal value
      -     > 0:  if info = i, U(i,i) computed in REAL is
                  exactly zero.  The factorization has been completed,
                  but the factor U is exactly singular, so the solution
                  could not be computed.

    @ingroup magma_sgesv_driver
    ********************************************************************/
extern "C" magma_int_t
magma_sgerfs_nopiv_mic(
    magma_trans_t trans, magma_int_t n, magma_int_t nrhs,
    magmaFloat_ptr dA, magma_int_t ldda,
    magmaFloat_ptr dB, magma_int_t lddb,
    magmaFloat_ptr dX, magma_int_t lddx,
    magmaFloat_ptr dworkd, magmaFloat_ptr dAF,
    magma_int_t *iter,
    magma_int_t *info, magma_queue_t queue)
{
    #define dB(i,j)     (dB + (i) + (j)*lddb)
    #define dX(i,j)     (dX + (i) + (j)*lddx)
    #define dR(i,j)     (dR + (i) + (j)*lddr)
    
    float c_neg_one = MAGMA_S_NEG_ONE;
    float c_one     = MAGMA_S_ONE;
    magma_int_t     ione  = 1;
    magmaFloat_ptr dR;
    magmaFloat_ptr Xnrmv, Rnrmv;
    float          Anrm, Xnrm, Rnrm, cte, eps;
    magma_int_t     i, j, iiter, lddsa, lddr;
    
    /* Check arguments */
    *iter = 0;
    *info = 0;
    if ( n < 0 )
        *info = -1;
    else if ( nrhs < 0 )
        *info = -2;
    else if ( ldda < max(1,n))
        *info = -4;
    else if ( lddb < max(1,n))
        *info = -8;
    else if ( lddx < max(1,n))
        *info = -10;
    
    if (*info != 0) {
        magma_xerbla( __func__, -(*info) );
        return *info;
    }
    
    if ( n == 0 || nrhs == 0 )
        return *info;

    lddsa = n;
    lddr  = n;
    
    dR  = dworkd;
    
    magma_smalloc_pinned(&Xnrmv, 1);
    magma_smalloc_pinned(&Rnrmv, 1);
    
    eps  = lapackf77_slamch("Epsilon");
    Anrm = magmablas_slange(MagmaInfNorm, n, n, dA,0, ldda, (float*)dworkd, queue );
    cte  = Anrm * eps * pow((float)n, 0.5) * BWDMAX;
    
       // residual dR = dB - dA*dX in real
    magmablas_slacpy( MagmaUpperLower, n, nrhs, dB, lddb, dR, lddr, queue );
    if ( nrhs == 1 ) {
        magma_sgemv( trans, n, n,
                     c_neg_one, dA, 0, ldda,
                                dX, 0, 1,
                     c_one,     dR, 0, 1, queue);
    }
    else {
        magma_sgemm( trans, MagmaNoTrans, n, nrhs, n,
                     c_neg_one, dA, 0, ldda,
                                dX, 0, lddx,
                     c_one,     dR, 0, lddr, queue );
    }
    
    // TODO: use MAGMA_S_ABS( dX(i,j) ) instead of slange?
    for( j=0; j < nrhs; j++ ) {
        i = magma_isamax( n, dX(0,j), 1, queue) - 1;
        magma_sgetmatrix( 1, 1, dX(i,j), 0, 1, Xnrmv, 0, 1, queue );
        Xnrm = lapackf77_slange( "F", &ione, &ione, Xnrmv, &ione, NULL );
        
        i = magma_isamax ( n, dR(0,j), 1, queue ) - 1;
        magma_sgetmatrix( 1, 1, dR(i,j), 0, 1, Rnrmv, 0, 1, queue );
        Rnrm = lapackf77_slange( "F", &ione, &ione, Rnrmv, &ione, NULL );

        if ( Rnrm >  Xnrm*cte ) {
            goto REFINEMENT;
        }
    }
    
    *iter = 0;
    return *info;

REFINEMENT:
    for( iiter=1; iiter < ITERMAX; ) {
        *info = 0;
        // solve dAF*dX = dR 
        // it's okay that dR is used for both dB input and dX output.
        magma_sgetrs_nopiv_mic( trans, n, nrhs, dAF, 0, lddsa, dR, 0, lddr, info, queue );
        if (*info != 0) {
            *iter = -3;
            goto FALLBACK;
        }
        
        // Add correction and setup residual
        // dX += dR  --and--
        // dR = dB
        // This saves going through dR a second time (if done with one more kernel).
        // -- not really: first time is read, second time is write.
        for( j=0; j < nrhs; j++ ) {
            magmablas_saxpycp2( n, dR(0,j), dX(0,j), dB(0,j), queue );
        }
        
        // residual dR = dB - dA*dX in real
        if ( nrhs == 1 ) {
            magma_sgemv( trans, n, n,
                         c_neg_one, dA, 0, ldda,
                                    dX, 0, 1,
                         c_one,     dR, 0, 1, queue );
        }
        else {
            magma_sgemm( trans, MagmaNoTrans, n, nrhs, n,
                         c_neg_one, dA, 0, ldda,
                                    dX, 0, lddx,
                         c_one,     dR, 0, lddr, queue );
        }
        
        /*  Check whether the nrhs normwise backward errors satisfy the
         *  stopping criterion. If yes, set ITER=IITER > 0 and return. */
        for( j=0; j < nrhs; j++ ) {
            i = magma_isamax( n, dX(0,j), 1, queue) - 1;
            magma_sgetmatrix( 1, 1, dX(i,j), 0, 1, Xnrmv, 0, 1 , queue);
            Xnrm = lapackf77_slange( "F", &ione, &ione, Xnrmv, &ione, NULL );
            
            i = magma_isamax ( n, dR(0,j), 1, queue ) - 1;
            magma_sgetmatrix( 1, 1, dR(i,j), 0, 1, Rnrmv,0, 1, queue );
            Rnrm = lapackf77_slange( "F", &ione, &ione, Rnrmv, &ione, NULL );
            
            if ( Rnrm >  Xnrm*cte ) {
                goto L20;
            }
        }
        
        /*  If we are here, the nrhs normwise backward errors satisfy
         *  the stopping criterion, we are good to exit. */
        *iter = iiter;
        return *info;
        
      L20:
        iiter++;
    }

    
    /* If we are at this place of the code, this is because we have
     * performed ITER=ITERMAX iterations and never satisified the
     * stopping criterion. Set up the ITER flag accordingly. */
    *iter = -ITERMAX - 1;
    
FALLBACK:
    /* Iterative refinement failed to converge to a
     * satisfactory solution. */
   
    free(Xnrmv);
    free(Rnrmv);

    return *info;
}




/**
    Purpose
    -------
    Solves a system of linear equations
       A * X = B
    where A is a general N-by-N matrix and X and B are N-by-nrhs matrices.
    Random Butterfly Tranformation is applied on A and B, then 
    the LU decomposition with no pivoting is
    used to factor A as
       A = L * U,
    where L is unit lower triangular, and U is
    upper triangular.  The factored form of A is then used to solve the
    system of equations A * X = B.
    The solution can then be improved using iterative refinement.


    Arguments
    ---------
    @param[in]
    ref  magma_bool_t
    Specifies if iterative refinement have to be applied to improve the solution.
    -     = MagmaTrue:   Iterative refinement is applied.
    -     = MagmaFalse:  Iterative refinement is not applied.

    @param[in]
    n       INTEGER
    The order of the matrix A.  N >= 0.

    @param[in]
    nrhs    INTEGER
    The number of right hand sides, i.e., the number of columns
    of the matrix B.  nrhs >= 0.

    @param[in,out]
    A       REAL array, dimension (LDA,N).
    On entry, the M-by-N matrix to be factored.
    On exit, the factors L and U from the factorization
    A = P*L*U; the unit diagonal elements of L are not stored.

    @param[in]
    lda     INTEGER
    The leading dimension of the array A.  LDA >= max(1,N).


    @param[in,out]
    B       REAL array, dimension (LDB,nrhs)
    On entry, the right hand side matrix B.
    On exit, the solution matrix X.

    @param[in]
    ldb     INTEGER
    The leading dimension of the array B.  LDB >= max(1,N).

    @param[out]
    info    INTEGER
    -     = 0:  successful exit
    -     < 0:  if INFO = -i, the i-th argument had an illegal value

    @ingroup magma_sgesv_driver
 ********************************************************************/




extern "C" magma_int_t 
magma_sgesv_rbt(
    magma_bool_t ref, magma_int_t n, magma_int_t nrhs, 
    float *A, magma_int_t lda, 
    float *B, magma_int_t ldb, 
    magma_int_t *info, magma_queue_t queue)
{

    /* Function Body */
    *info = 0;
    if (n < 0) {
        *info = -1;
    } else if (nrhs < 0) {
        *info = -2;
    } else if (lda < max(1,n)) {
        *info = -4;
    } else if (ldb < max(1,n)) {
        *info = -6;
    }
    if (*info != 0) {
        magma_xerbla( __func__, -(*info) );

        return *info;
    }

    /* Quick return if possible */
    if (nrhs == 0 || n == 0)
        return *info;


    magma_int_t nn = n + ((4-(n % 4))%4);
    float *dA, *hu, *hv, *db, *dAo, *dBo, *dwork;

    magma_int_t n2;

    n2 = nn*nn;

    if (MAGMA_SUCCESS != magma_smalloc( &dA, n2 )) {
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }
    if (MAGMA_SUCCESS != magma_smalloc( &db, nn*nrhs )) {
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }

    if (ref == MagmaTrue) {
        if (MAGMA_SUCCESS != magma_smalloc( &dAo, n2 )) {
            *info = MAGMA_ERR_DEVICE_ALLOC;
            return *info;
        }

        if (MAGMA_SUCCESS != magma_smalloc( &dwork, nn*nrhs )) {
            *info = MAGMA_ERR_DEVICE_ALLOC;
            return *info;
        }
        if (MAGMA_SUCCESS != magma_smalloc( &dBo, nn*nrhs )) {
            *info = MAGMA_ERR_DEVICE_ALLOC;
            return *info;
        }
    }

    
    magma_malloc_pinned((void**)&hu, 2*nn*sizeof(float ));
    magma_malloc_pinned((void**)&hv, 2*nn*sizeof(float ));
    
    magma_slaset(MagmaFull, nn, nn, MAGMA_S_ZERO, MAGMA_S_ONE, dA, 0, nn, queue);

    /* Send matrix on the MIC*/
    magma_ssetmatrix(n, n, A, 0, lda, dA, 0, nn, queue);

    /* Send b on the MIC*/
    magma_ssetmatrix(n, nrhs, B, 0, ldb, db, 0, nn, queue);

    magma_int_t iter;

    *info = magma_sgerbt_mic(MagmaTrue, nn, nrhs, dA, nn, db, nn, hu, hv, info, queue);
    if(*info != MAGMA_SUCCESS)  {
        return *info;
    }
    

    if (ref == MagmaTrue) {
        magmablas_slacpy(MagmaFull, nn, nn, dA, nn, dAo, nn, queue);
        magmablas_slacpy(MagmaFull, nn, nrhs, db, nn, dBo, nn, queue);
    }

    /* Solve the system U^TAV.y = U^T.b on the MIC*/ 
    magma_sgesv_nopiv_mic( nn, nrhs, dA, 0, nn, db,0, nn, info, queue);

   if (ref == MagmaTrue)
    magma_sgerfs_nopiv_mic(MagmaNoTrans, nn, nrhs, dAo, nn, dBo, nn, db, nn, dwork, dA, &iter, info, queue);




    /* The solution of A.x = b is Vy computed on the MIC */
    float *dv;

    if (MAGMA_SUCCESS != magma_smalloc( &dv, 2*nn )) {
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }

    magma_ssetvector(2*nn, hv, 1, dv, 1, queue);


    for(int i = 0; i < nrhs; i++)
        magma_sprbt_mv(nn, dv, db+(i*nn), queue);

    magma_sgetmatrix(n, nrhs, db, 0, nn, B, 0, ldb, queue);

    free( hu);
    free( hv);

    magma_free( dA );
    magma_free( dv );
    magma_free( db );
    if (ref == MagmaTrue) {    
        magma_free( dAo );
        magma_free( dBo );
        magma_free( dwork );
    }

    return(*info);
}


