/*
 *  -- MAGMA (version 1.4.0) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     @date July 2015
 *
 * @generated from testing_zgetrf_mmic.cpp normal z -> c, Sun Jul 12 17:46:29 2015
 * @author Khairul Kabir
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

#define PRECISION_c

// Initialize matrix to random.
// Having this in separate function ensures the same ISEED is always used,
// so we can re-generate the identical matrix.
void init_matrix( magma_int_t m, magma_int_t n, magmaFloatComplex *h_A, magma_int_t lda )
{
    magma_int_t ione = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t n2 = lda*n;
    lapackf77_clarnv( &ione, ISEED, &n2, h_A );
}


// On input, A and ipiv is LU factorization of A. On output, A is overwritten.
// Requires m == n.
// Uses init_matrix() to re-generate original A as needed.
// Generates random RHS b and solves Ax=b.
// Returns residual, |Ax - b| / (n |A| |x|).
float get_residual(
    magma_int_t m, magma_int_t n,
    magmaFloatComplex *A, magma_int_t lda,
    magma_int_t *ipiv )
{
    if ( m != n ) {
        printf( "\nERROR: residual check defined only for square matrices\n" );
        return -1;
    }
    
    const magmaFloatComplex c_one     = MAGMA_C_ONE;
    const magmaFloatComplex c_neg_one = MAGMA_C_NEG_ONE;
    const magma_int_t ione = 1;
    
    // this seed should be DIFFERENT than used in init_matrix
    // (else x is column of A, so residual can be exactly zero)
    magma_int_t ISEED[4] = {0,0,0,2};
    magma_int_t info = 0;
    magmaFloatComplex *x, *b;
    
    // initialize RHS
    TESTING_MALLOC_HOST( x, magmaFloatComplex, n );
    TESTING_MALLOC_HOST( b, magmaFloatComplex, n );
    lapackf77_clarnv( &ione, ISEED, &n, b );
    blasf77_ccopy( &n, b, &ione, x, &ione );
    
    // solve Ax = b
    lapackf77_cgetrs( "Notrans", &n, &ione, A, &lda, ipiv, x, &n, &info );
    if (info != 0)
        printf("lapackf77_cgetrs returned error %d: %s.\n",
               (int) info, magma_strerror( info ));
    
    // reset to original A
    init_matrix( m, n, A, lda );
    
    // compute r = Ax - b, saved in b
    blasf77_cgemv( "Notrans", &m, &n, &c_one, A, &lda, x, &ione, &c_neg_one, b, &ione );
    
    // compute residual |Ax - b| / (n*|A|*|x|)
    float norm_x, norm_A, norm_r, work[1];
    norm_A = lapackf77_clange( "F", &m, &n, A, &lda, work );
    norm_r = lapackf77_clange( "F", &n, &ione, b, &n, work );
    norm_x = lapackf77_clange( "F", &n, &ione, x, &n, work );
    
    //printf( "r=\n" ); magma_cprint( 1, n, b, 1 );
    
    TESTING_FREE_HOST( x );
    TESTING_FREE_HOST( b );
    
    //printf( "r=%.2e, A=%.2e, x=%.2e, n=%d\n", norm_r, norm_A, norm_x, n );
    return norm_r / (n * norm_A * norm_x);
}


// On input, LU and ipiv is LU factorization of A. On output, LU is overwritten.
// Works for any m, n.
// Uses init_matrix() to re-generate original A as needed.
// Returns error in factorization, |PA - LU| / (n |A|)
// This allocates 3 more matrices to store A, L, and U.
float get_LU_error(magma_int_t M, magma_int_t N,
                    magmaFloatComplex *LU, magma_int_t lda,
                    magma_int_t *ipiv)
{
    magma_int_t min_mn = min(M,N);
    magma_int_t ione   = 1;
    magma_int_t i, j;
    magmaFloatComplex alpha = MAGMA_C_ONE;
    magmaFloatComplex beta  = MAGMA_C_ZERO;
    magmaFloatComplex *A, *L, *U;
    float work[1], matnorm, residual;
    
    TESTING_MALLOC_HOST( A, magmaFloatComplex, lda*N    );
    TESTING_MALLOC_HOST( L, magmaFloatComplex, M*min_mn );
    TESTING_MALLOC_HOST( U, magmaFloatComplex, min_mn*N );
    memset( L, 0, M*min_mn*sizeof(magmaFloatComplex) );
    memset( U, 0, min_mn*N*sizeof(magmaFloatComplex) );

    // set to original A
    init_matrix( M, N, A, lda );
    lapackf77_claswp( &N, A, &lda, &ione, &min_mn, ipiv, &ione);
    
    // copy LU to L and U, and set diagonal to 1
    lapackf77_clacpy( MagmaLowerStr, &M, &min_mn, LU, &lda, L, &M      );
    lapackf77_clacpy( MagmaUpperStr, &min_mn, &N, LU, &lda, U, &min_mn );
    for(j=0; j<min_mn; j++)
        L[j+j*M] = MAGMA_C_MAKE( 1., 0. );
    
    matnorm = lapackf77_clange("f", &M, &N, A, &lda, work);

    blasf77_cgemm("N", "N", &M, &N, &min_mn,
                  &alpha, L, &M, U, &min_mn, &beta, LU, &lda);

    for( j = 0; j < N; j++ ) {
        for( i = 0; i < M; i++ ) {
            LU[i+j*lda] = MAGMA_C_SUB( LU[i+j*lda], A[i+j*lda] );
        }
    }
    residual = lapackf77_clange("f", &M, &N, LU, &lda, work);

    TESTING_FREE_HOST( A );
    TESTING_FREE_HOST( L );
    TESTING_FREE_HOST( U );

    return residual / (matnorm * N);
}

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cgetrf
   */
int main( int argc, char** argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    magma_int_t num = 0;
    magma_int_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
        fprintf( stderr, "magma_get_devices failed: %d\n", (int) err );
        exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
        fprintf( stderr, "magma_queue_create failed: %d\n", (int) err );
        exit(-1);
    }

    real_Double_t   gflops, mic_perf, mic_time, cpu_perf, cpu_time;
    float          error;
    magmaFloatComplex *h_A;
    magmaFloatComplex *d_lA[4] = {NULL, NULL, NULL, NULL};
    magma_int_t     *ipiv;
    magma_int_t M, N, n2, lda, ldda, info, min_mn;
    magma_int_t num_mics, num_mics0 = 1, n_local;
    magma_int_t i, j, k, nb0, nb, nk;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t status   = 0;

    magma_opts opts;
    parse_opts( argc, argv, &opts );
    num_mics0 = opts.nmic;

    float tol = opts.tolerance * lapackf77_slamch("E");
    
    if ( opts.check == 2 ) {
        printf("    M     N   CPU GFlop/s (sec)   MIC GFlop/s (sec)   |Ax-b|/(N*|A|*|x|)\n");
    }
    else {
        printf("    M     N   CPU GFlop/s (sec)   MIC GFlop/s (sec)   |PA-LU|/(N*|A|)\n");
    }
    printf("=========================================================================\n");
 
    for( int itest = 0; itest < opts.ntest; ++itest ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            M = opts.msize[itest];
            N = opts.nsize[itest];
            min_mn = min(M, N);
            lda    = M;
            n2     = lda*N;
            ldda   = ((M+31)/32)*32;
            gflops = FLOPS_CGETRF( M, N ) / 1e9;

            TESTING_MALLOC_HOST(    ipiv, magma_int_t,     min_mn );
            TESTING_MALLOC_HOST(    h_A,  magmaFloatComplex, n2     );

            //TESTING_MALLOC_DEV (  d_A,  magmaFloatComplex, ldda*N );
            /* == distributing the matrix == */
            nb = magma_get_cgetrf_nb(N);
            
            if ( num_mics0 > N / nb ) {
                num_mics = N / nb;
                if ( N % nb != 0 ) num_mics ++;
                printf( " * too many MICs for the matrix size, using %d MICs\n", (int) num_mics );
            } else {
                num_mics = num_mics0;
            }

            /* allocate device memory, assuming fixed nb and num_mics */
            for(j = 0; j < num_mics; j++) {
                n_local = ((N / nb) / num_mics) * nb;
                if (j < ((N / nb) % num_mics))
                    n_local += nb;
                else if (j == (N / nb) % num_mics)
                    n_local += N % nb;
                magma_setdevice(j);

                d_lA[j] = 0;
                if (n_local > 0) {
                    TESTING_MALLOC_DEV( d_lA[j], magmaFloatComplex, ldda * n_local );
                }
            }

	    /* =====================================================================
               Performs operation using LAPACK
               =================================================================== */
            if ( opts.lapack ) {
                init_matrix( M, N, h_A, lda );
                
                cpu_time = magma_wtime();
                lapackf77_cgetrf(&M, &N, h_A, &lda, ipiv, &info);
                cpu_time = magma_wtime() - cpu_time;
                cpu_perf = gflops / cpu_time;
                if (info != 0)
                    printf("lapackf77_cgetrf returned error %d.\n",
                           (int) info);
            }
 
            /* ====================================================================
               Performs operation using MAGMA
               =================================================================== */
            init_matrix( M, N, h_A, lda );
            for( j = 0; j < N; j += nb) {
                k = (j / nb) % num_mics;
                magma_setdevice(k);
                nk = min(nb, N - j);
                magma_csetmatrix( M, nk,
                        h_A + j * lda, 0,                      lda,
                        d_lA[k] + j / (nb * num_mics) * nb * ldda, 0, ldda, queue);
            }

            magma_setdevice(0);

            /* == calling MAGMA with multiple MICs == */
            mic_time = magma_wtime();
            magma_cgetrf_mmic( num_mics, M, N, d_lA, ldda, ipiv, &info, queue);
            mic_time = magma_wtime() - mic_time;
            mic_perf = gflops / mic_time;
            if (info < 0) {
                printf("Argument %d of magma_cgetrf_mmic had an illegal value.\n", (int) -info);
                break;
            } else if (info != 0 ) {
                printf("magma_cgetrf_mmic returned info=%d.\n", (int) info);
                break;
            }

            /* =====================================================================
               Check the factorization
               =================================================================== */
            if ( opts.lapack ) {
                printf("%5d %5d   %7.2f (%7.2f)   %7.2f (%7.2f)",
                       (int) M, (int) N, cpu_perf, cpu_time, mic_perf, mic_time );
            }
            else {
                printf("%5d %5d     ---   (  ---  )   %7.2f (%7.2f)",
                       (int) M, (int) N, mic_perf, mic_time );
            }
            if ( opts.check == 2 ) {
                /* == download the matrix from MICs == */
                for(j = 0; j < N; j += nb) {
                    k = (j / nb) % num_mics;
                    magma_setdevice(k);
                    nk = min(nb, N - j);
                    magma_cgetmatrix( M, nk,
                            d_lA[k]+j/(nb*num_mics)*nb*ldda, 0, ldda,
                            h_A+j*lda,  0,                   lda, queue );
                }

                magma_setdevice(0);

                error = get_residual( M, N, h_A, lda, ipiv );
                printf("   %8.2e  %s\n", error, (error < tol ? "ok" : "failed"));
                status |= ! (error < tol);
            }
            else if ( opts.check ) {
                /* == download the matrix from MICs == */
                for(j = 0; j < N; j += nb) {
                    k = (j / nb) % num_mics;
                    magma_setdevice(k);
                    nk = min(nb, N - j);
                    magma_cgetmatrix( M, nk,
                            d_lA[k]+j/(nb*num_mics)*nb*ldda, 0, ldda,
                            h_A+j*lda,  0,                   lda, queue );
                }
                magma_setdevice(0);

                error = get_LU_error( M, N, h_A, lda, ipiv );
                printf("   %8.2e  %s\n", error, (error < tol ? "ok" : "failed"));
                status |= ! (error < tol);
            }
            else {
                printf("     ---  \n");
            }
 
            fflush(0);
            /* Memory clean up */
            TESTING_FREE_HOST( ipiv );
            TESTING_FREE_HOST( h_A );
            //TESTING_FREE_DEV ( d_A );
            for(j = 0; j < num_mics0; j++) {
                magma_setdevice(j);
                if (d_lA[j] != NULL)
                    TESTING_FREE_DEV( d_lA[j] );
            }


        }
        if ( opts.niter > 1 ) {
            printf( "\n" );
        }
    }


    /* Shutdown */
    magma_queue_destroy(queue);
    magma_finalize();
    return 0;
}
